from typing import Any
from sqlalchemy import Column, Integer, REAL
from .bmp_table import BMPTable
from ...names import Names

class SubbasinMultiplier(BMPTable):
    __tablename__ = Names.bmp_table_name_subbasin_multiplier

    ID = Column(Integer, primary_key=True)
    SHC = Column(REAL)
    SOIL_T10 = Column(REAL)
    INTERC_MAX = Column(REAL)
    INTERC_MIN = Column(REAL)
    USLE_C = Column(REAL)
    ROOT_DEPTH = Column(REAL)
    MANNING = Column(REAL)
    FIMP = Column(REAL)
    FCIMP = Column(REAL)
    CURBDEN = Column(REAL)
    URBCOEF = Column(REAL)
    DIRTMX = Column(REAL)
    THALF = Column(REAL)
    TNCONC = Column(REAL)
    TPCONC = Column(REAL)
    TNO3CONC = Column(REAL)
    URBCN2 = Column(REAL)
    PET_LAI_COEFFICIENT = Column(REAL)
    CONDUCTIVITY = Column(REAL)
    DEPTH = Column(REAL)
    POROSITY = Column(REAL)
    FIELD_CAPACITY = Column(REAL)
    WILTING_POINT = Column(REAL)
    RESIDUAL = Column(REAL)
    PORE_INDEX = Column(REAL)
    DENSITY = Column(REAL)
    SAND = Column(REAL)
    CLAY = Column(REAL)
    SLIT = Column(REAL)
    USLE_K = Column(REAL)
    CN2 = Column(REAL)
    RUNOFF_CO = Column(REAL)
    DEPRESSION = Column(REAL)
    SOIL_MOIST_IN = Column(REAL)
    SOIL_NO3_IN = Column(REAL)
    SOIL_ORGN_IN = Column(REAL)
    SOIL_SOLP_IN = Column(REAL)
    SOIL_ORGP_IN = Column(REAL)
    INTERFLOW_SCALE_FACTOR = Column(REAL)

    def __init__(self, id:int):
        self.ID = id
        self.SHC = 1
        self.SOIL_T10 = 1
        self.INTERC_MAX = 1
        self.INTERC_MIN = 1
        self.USLE_C = 1
        self.ROOT_DEPTH = 1
        self.MANNING = 1
        self.FIMP = 1
        self.FCIMP = 1
        self.CURBDEN = 1
        self.URBCOEF = 1
        self.DIRTMX = 1
        self.THALF = 1
        self.TNCONC = 1
        self.TPCONC = 1
        self.TNO3CONC = 1
        self.URBCN2 = 1
        self.PET_LAI_COEFFICIENT = 1
        self.CONDUCTIVITY = 1
        self.DEPTH = 1
        self.POROSITY = 1
        self.FIELD_CAPACITY = 1
        self.WILTING_POINT = 1
        self.RESIDUAL = 1
        self.PORE_INDEX = 1
        self.DENSITY = 1
        self.SAND = 1
        self.CLAY = 1
        self.SLIT = 1
        self.USLE_K = 1
        self.CN2 = 1
        self.RUNOFF_CO = 1
        self.DEPRESSION = 1
        self.SOIL_MOIST_IN = 1
        self.SOIL_NO3_IN = 1
        self.SOIL_ORGN_IN = 1
        self.SOIL_SOLP_IN = 1
        self.SOIL_ORGP_IN = 1
        self.INTERFLOW_SCALE_FACTOR = 1




from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from .bmp_management_base import BMPManagementBaseWithYear
from ...names import Names

class ManureStorageParameter(BMPTable):
    """Parameter Table for BMP: Manure storage capacity and design (27)"""
    __tablename__ = Names.bmp_table_name_manure_storage_parameter
    """Manure storage ID"""    
    ID = Column(Integer,primary_key=True)
    """Name"""
    Name = Column(TEXT)
    """Description"""
    Description = Column(TEXT)
    """Producer ID"""
    ProducerID = Column(Integer)
    """Subbasin ID"""
    Subbasin = Column(Integer)
    """Corresponding feedlot ID"""
    FeedlotID = Column(Integer)
    """Length of the manure storage (m)"""
    ManLength_m = Column(REAL)
    """Width of the Manure storage (m)"""
    ManWidth_m = Column(REAL)
    """Area of the Manure storage (m2)"""
    ManArea_m2 = Column(REAL)
    """Drainage area of the manure storage (ha)"""
    DraArea_ha = Column(REAL)
    """Manure drainage area fraction of the subbasin area (-)"""
    DraFraction = Column(REAL)
    """Distance to the nearby stream"""
    DisReach_m = Column(REAL)
    """Height to the nearby stream"""
    HigReach_m = Column(REAL)
    """Threshold value of storage distance to the nearby stream (100m)"""
    ThDisReach = Column(REAL)
    """Threshold value of storage height to the nearby stream (10m)"""
    ThHigReach = Column(REAL)
    """Initial storage (kg)"""
    ManInitial = Column(REAL)
    """CN change fraction compared to non-storage area"""
    CN_change = Column(REAL)
    """PRC change fraction compared to non-storage area"""
    PRC_change = Column(REAL)
    """Manure event mean concentration (mg/l)"""
    Manure_EMC = Column(REAL)

    def __init__(self, 
                 id:int, 
                 subbasin:int, 
                 feedlot:int, 
                 contribution_area_ha:float, 
                 drainage_fraction:float, 
                 distance_to_reach:float, 
                 height_to_reach:float):
        
        self.ID = id
        self.ProducerID = id
        self.Subbasin = subbasin
        self.FeedlotID = feedlot
        self.DraArea_ha = contribution_area_ha
        self.DraFraction = drainage_fraction
        self.DisReach_m = distance_to_reach
        self.HigReach_m = height_to_reach

        self.ManLength_m = 20
        self.ManWidth_m = 20
        self.ManArea_m2 = 400

        self.ThDisReach = 0
        self.ThHigReach = 0

        self.ManInitial = 0
        self.CN_change = 0.3
        self.PRC_change = 0.3
        self.Manure_EMC = 10000

class ManureStorageManagement(BMPManagementBaseWithYear):
    def __init__(self):
        super().__init__()
        """Month of manure application"""
        self.ManAppMon = 11
        """Day of manure application"""
        self.ManAppDay = 15
        """Fraction of manure amount applied to the field"""
        self.ManAppFra = 1





from .bmp_management_base import BMPManagementBase

class ManureNoApplicaitonOnSnowManagement(BMPManagementBase):
    """Distribution Table for BMP: No application on snow (23)"""
    def __init__(self):
        super().__init__()



from sqlalchemy import Column, Integer, Float, String, Double
from .bmp_table import BMPTable
from ...names import Names

class BMP_index(BMPTable):
    __tablename__ = Names.bmp_table_name_bmp_index
    ID = Column(Integer)
    Name = Column(String)
    Type = Column(Integer)
    CODE = Column(String)


import os
from whitebox_workflows import Raster, Vector
from .parameters import Parameters
from .database.bmp.reach import Reach
import numpy as np
import math
from .folder_base import FolderBase
from .names import Names
from .inputs import Inputs
from .delineation.structure import Structure
from .raster_extension import RasterExtension
from .delineation.delineation import Delineation
from .vector_extension import VectorExtension
from .database.parameter.parameter_database import ParameterDatabase
import pandas as pd
from .bmp.bmp_type import DefaultScenarioId

import logging
logger = logging.getLogger(__name__)
        
class Outputs(FolderBase):
    """
    Output folder where all intermediate and final processed files are saved
    """
    def __init__(self, 
                 output_folder:str, 
                 input_folder:str,
                 database_folder:str, 
                 stream_threshold_area_ha:float = 10,   #stream thrshold area
                 wetland_min_area_ha:float = 0.1,       #min wetland area
                 design_storm_return_period = 2         #design storm return period for reach width and depth
                 ) -> None:
        super().__init__(output_folder)

        self.inputs = Inputs(input_folder)       
        self.database_folder = database_folder 

        self.stream_threshold_area_ha = stream_threshold_area_ha
        self.main_stream_threshold_area_ha = 200            #threshold for main stream in ha
        self.design_storm_return_period = design_storm_return_period

        self.wetland_min_area_ha = wetland_min_area_ha
        self.__structures = None
        self.__structure_combined = None
        self.__parameters = None

    @property
    def number_of_valid_cell(self)->int:
        return RasterExtension.get_number_of_valid_cell(self.mask_refined_with_subbasin_raster)
    
    @property
    def number_of_subbasin(self)->int:
        return int(RasterExtension.get_max_value(self.subbasin_raster))
    
    @property
    def parameter(self)->Parameters:
        if self.__parameters is None:
            self.__parameters = Parameters(self.database_folder)

        return self.__parameters
    
#region structures

    @property
    def wetland_raster(self)->Raster:
        """
        create wetland raster
        1. if user provides wetland vector, use it
        2. if not, create an empty raster as it must have one
        """
        raster = self.get_raster(Names.wetlandRasName)

        if raster is None:
            if "wetland" in self.structures:
                raster = self.structures["wetland"].boundary_raster
            else:
                raster = self.inputs.create_new_raster()
            
            raster = self.save_raster(raster, Names.wetlandRasName, True, True)

        return raster

    @property
    def structures(self)->dict[str, Structure]:
        if self.__structures is None:
            self.__structures = {}
            logger.info("Initilizing structures ...")
            for type in Structure.structure_types_affection_flow_direction:
                boundary_vector = getattr(self.inputs, f"{type}_boundary_vector")
                if boundary_vector is not None:
                    logger.info(type)
                    self.__structures[type] = Structure(
                        structure_type=type,
                        output_folder=self.folder,
                        dem_raster=self.dem_clipped_burned_filled_raster,
                        structure_polygon_vector = boundary_vector, 
                        structure_outlet_point_vector = getattr(self.inputs, f"{type}_outlet_vector"),
                        structure_area_threshold_ha = self.wetland_min_area_ha if type == "wetland" else 0)     
                    
        return self.__structures
    
    @property
    def structure_combined_boundary_vector(self)->Vector:
        vector = self.get_vector(Names.structureCombinedBoundaryShpName)

        if vector is None and len(self.structures) > 0:
            vector = VectorExtension.merge_vectors([structure.boundary_original_vector for structure in self.structures.values()])
            self.save_vector(vector, Names.structureCombinedBoundaryShpName)

        return vector
    
    @property
    def structure_combined_outlet_vector(self)->Vector:
        vector = self.get_vector(Names.structureCombinedOutputShpName)

        if vector is None and len(self.structures) > 0:
            vector = VectorExtension.merge_vectors([structure.outlet_vector for structure in self.structures.values()])
            self.save_vector(vector, Names.structureCombinedOutputShpName)

        return vector
    
    @property
    def structure_combined_outlet_raster(self):
        raster = self.get_raster(Names.structureCombinedOutputRasName)

        if raster is None and len(self.structures) > 0:
            raster = self.wbe.vector_points_to_raster(input = self.structure_combined_outlet_vector, 
                                                            field_name = Names.field_name_id,
                                                            base_raster = self.inputs.dem_raster)
            raster = self.save_raster(raster, Names.structureCombinedOutputRasName, True, True)

        return raster

    @property
    def structure_combined(self)->Structure:
        """
        Combined structure used to get the revised flow direction
        """
        if len(self.structures) <= 0:
            return None
        
        if self.__structure_combined is None:
            logger.info("Merging all flow-direction-affecting structures ...")

            self.__structure_combined = Structure(
                structure_type="structureCombined",
                output_folder=self.folder,
                dem_raster=self.dem_clipped_burned_filled_raster,
                structure_polygon_vector = self.structure_combined_boundary_vector,
                structure_outlet_point_vector = self.structure_combined_outlet_vector
            )

        return self.__structure_combined

#endregion
                
#region DEM processing

    @property
    def mask_raster(self):
        raster = self.get_raster(Names.maskRasName)
 
        if raster is None:
            #try user provided boundary shapefile
            if self.inputs.boundary_vector is not None:
                logger.info(f"Creating mask using user-defined boundary shapefile ... ")
                raster = self.wbe.vector_polygons_to_raster(input = self.inputs.boundary_vector, 
                                                            base_raster = self.inputs.dem_raster)
                raster = raster.con(f"value == {raster.configs.nodata}", self.inputs.nodata, 1)                
            #try use soil and landuse 
            elif self.inputs.soil_raster is not None and self.inputs.landuse_raster is not None:
                logger.info(f"Creating mask using soil and landuse ... ")
                raster = self.inputs.create_new_raster()
                raster = self.inputs.dem_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, 1)
                raster = self.inputs.soil_raster.con(f"value == {self.inputs.soil_raster.configs.nodata}", self.inputs.nodata, raster)
                raster = self.inputs.landuse_raster.con(f"value == {self.inputs.landuse_raster.configs.nodata}", self.inputs.nodata, raster)

            if raster is not None:
                raster = self.save_raster(raster, Names.maskRasName, True, True)

        return raster
    
    @property
    def mask_refined_with_subbasin_raster(self):
        raster = self.get_raster(Names.maskRefindedWithSubbasinRasName)

        if raster is None:
            logger.info("Refining mask with subbain ...")
            raster = self.subbasin_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, 1)
            raster = self.inputs.soil_raster.con(f"value == {self.inputs.soil_raster.configs.nodata}", self.inputs.nodata, raster)
            raster = self.inputs.landuse_raster.con(f"value == {self.inputs.landuse_raster.configs.nodata}", self.inputs.nodata, raster)

            raster = self.save_raster(raster, Names.maskRefindedWithSubbasinRasName, True, True)

        return raster

    @property 
    def dem_clipped_raster_for_model(self):
        """
        This is the dem clipped by the final subbain raster. This will be the base for parameter.h5
        """
        raster = self.get_raster(Names.demName)

        if raster is None:
            raster = self.mask_refined_with_subbasin_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, self.inputs.dem_raster)
            raster = self.save_raster(raster, Names.demName, True, True)

        return raster

    @property 
    def dem_clipped_raster(self):
        """
        Clipped DEM by mask
        """
        raster = self.get_raster(Names.demClippedName)

        if raster is None:
            logger.info("Masking DEM ...")
            raster = self.mask_raster.con("value == 1", self.inputs.dem_raster, self.inputs.nodata)
            raster = self.save_raster(raster, Names.demClippedName, True, True)

        return raster

    @property 
    def dem_clipped_burned_raster(self):
        """
        Burned DEM with user-provided stream shapefile
        """
        if self.stream_network_user_raster is None:
            return self.dem_clipped_raster

        raster = self.get_raster(Names.demBurnedName)
        if raster is None:
            logger.info("Burning stream ...")
            raster = self.stream_network_user_raster.con(f"value == nodata",
                                                         self.dem_clipped_raster,
                                                         self.dem_clipped_raster - 10)
            raster = self.save_raster(raster, Names.demBurnedName, True, True)

        return raster
    
    @property
    def dem_clipped_burned_filled_raster(self):
        """
        Filled DEM
        """
        raster = self.get_raster(Names.demFilledName)

        if raster is None:
            logger.info("Filling depression ...")
            #raster = self.wbe.breach_depressions_least_cost(self.dem_clipped_burned_raster, flat_increment=0.001, max_dist=100) # Change the max dist parameter as appropriate for your DEM
            raster = self.wbe.fill_depressions(self.dem_clipped_burned_raster, flat_increment=0.001)
            raster = self.save_raster(raster, Names.demFilledName, True, True)

        return raster

#endregion
    
#region farm & field

    # def __remove_non_agriculture_cell(self, raster:Raster)->Raster:
    #     """
    #     remove non agriculture cells

    #     Replace filterAgriLanduse
    #     """
    #     raster_with_only_algriculture = raster.farm_raster.deep_copy()
    #     algricultrual_landuses = self.parameter.bmp_database.algricultural_landuses

    #     for row in raster.configs.rows:
    #         for col in raster.configs.columns:
    #             landuse = self.mapped_landuse_raster[row, col]
                
    #             if landuse > 0 and landuse not in algricultrual_landuses:
    #                 raster_with_only_algriculture[row, col] = raster.configs.nodata

    #     return raster_with_only_algriculture

    @property
    def farm_raster(self)->Raster:
        raster = self.get_raster(Names.farmRasName)
    
        if raster is None:
            if self.inputs.farm_vector is not None:
                exist, id_field_name = VectorExtension.check_id(self.inputs.farm_vector)
                if not exist:
                    raise ValueError(f"Couldn't find ID column in {self.inputs.farm_vector.file_name} ...")
                raster = self.wbe.vector_polygons_to_raster(input = self.inputs.farm_vector, 
                                                            field_name = id_field_name,
                                                            base_raster = self.inputs.dem_raster)
                raster = self.mask_refined_with_subbasin_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, raster)
                raster = self.save_raster(raster, Names.farmRasName, True, True)

        return raster
    
    # @property
    # def farm_without_agriculture_raster(self)->Raster:
    #     raster = self.get_raster(Names.farmWithOnlyAgricultureRasName)

    #     if raster is None:
    #         raster = self.__remove_non_agriculture_cell(self.farm_raster)
    #         raster = self.mask_refined_with_subbasin_raster.con("value == 1", raster, raster.configs.nodata)
    #         self.save_raster(raster, Names.farmWithOnlyAgricultureRasName)

    #     return raster

    
    @property
    def field_raster(self)->Raster:
        raster = self.get_raster(Names.fieldRasName)
    
        if raster is None:
            if self.inputs.field_vector is not None:
                exist, id_field_name = VectorExtension.check_id(self.inputs.field_vector)
                if not exist:
                    raise ValueError(f"Couldn't find ID column in {self.inputs.field_vector.file_name} ...")
                raster = self.wbe.vector_polygons_to_raster(
                    input = self.inputs.field_vector, 
                    field_name = id_field_name,
                    base_raster = self.inputs.dem_raster)
                raster = self.mask_refined_with_subbasin_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, raster)
                raster = self.save_raster(raster, Names.fieldRasName, True, True)

        return raster
    
    @property
    def field_clipped_vector(self)->Vector:
        """Clipped field vector by subbasin"""
        vector = self.get_vector(Names.fieldClippedShpName)

        if vector is None:
            if self.field_raster is not None:
                vector = RasterExtension.raster_to_vector(self.field_raster)
                self.save_vector(vector, Names.fieldClippedShpName)

        return vector

        
    
    # @property
    # def field_without_agriculture_raster(self)->Raster:
    #     raster = self.get_raster(Names.fieldWithOnlyAgricultureRasName)

    #     if raster is None:
    #         raster = self.__remove_non_agriculture_cell(self.field_raster)
    #         raster = self.mask_refined_with_subbasin_raster.con("value == 1", raster, raster.configs.nodata)
    #         self.save_raster(raster, Names.fieldWithOnlyAgricultureRasName)

    #     return raster

    @property
    def subarea_raster(self)->Raster:
        raster = self.get_raster(Names.subareaRasName)

        if raster is None:
            if self.field_raster is not None:
                raster, raster1_max = RasterExtension.get_overlay_raster(self.field_raster, self.subbasin_raster)
                raster = self.save_raster(raster, Names.subareaRasName, True, True)

                #convert to vector and assign field and subbasin id
                vector = RasterExtension.raster_to_vector(raster)
                vector = VectorExtension.decompsite_overlay_id(vector, "field", "subbasin", raster1_max)
                self.save_vector(vector, Names.subareaRasName)

        return raster


    @property
    def subarea_vector(self)->Raster:
        vector = self.get_vector(Names.subareaShpName)

        if vector is None:
            raster = self.subarea_raster
            vector = self.get_vector(Names.subareaShpName)

        return vector

#endregion    

#region parameters

    @property
    def mapped_soil_raster(self)->Raster:
        """
        mapped soil raster based on lookup table
        The name must be soil as specified in the engine
        """
        raster = self.get_raster(Names.soilName)

        if raster is None:
            raster = self.inputs.lookup_soil.mapped_raster
            raster = self.save_raster(raster, Names.soilName, True, True)

        return raster
    
    @property
    def soil_k_raster(self)->Raster:
        return RasterExtension.reclassify(self.mapped_soil_raster, self.parameter.get_parameter_lookup("AverageK","soil"), self.subbasin_raster)

    @property
    def soil_porosity_raster(self)->Raster:
        return RasterExtension.reclassify(self.mapped_soil_raster, self.parameter.get_parameter_lookup("AveragePorosity","soil"), self.subbasin_raster)

    @property
    def landuse_rootdepth_raster(self)->Raster:
        return RasterExtension.reclassify(self.mapped_landuse_raster, self.parameter.get_parameter_lookup("ROOT_DEPTH","landuse"), self.subbasin_raster)

    @property
    def mapped_landuse_raster(self)->Raster:
        """
        mapped landuse raster based on lookup table. 
        The name must be landuse as specified in the engine
        """
        raster = self.get_raster(Names.landuseName)

        if raster is None:
            raster = self.inputs.lookup_landuse.mapped_raster
            raster = self.save_raster(raster, Names.landuseName, True, True)

        return raster

    @property
    def uslep_raster(self)->Raster:
        """
        use default usle P as 1, just copy from the mask
        """
        raster = self.get_raster(Names.uslePName)
        if raster is None:
            logger.info("Creating USLE P raster ...")
            self.save_raster(self.mask_refined_with_subbasin_raster, Names.uslePName)
            raster = self.get_raster(Names.uslePName)

        return raster

    @property
    def field_capacity_raster(self)->Raster:
        """
        field capacity from soil and soil parameter table in parameter database
        """
        raster = self.get_raster(Names.fieldCapName)
        if raster is None:
            raster = RasterExtension.reclassify(self.mapped_soil_raster, self.parameter.get_parameter_lookup("FC1","soil"), self.subbasin_raster)
            raster = self.save_raster(raster, Names.fieldCapName, True, True)      

        return raster
    
    @property
    def manning_raster(self)->Raster:
        """
        manning from landuse and landuse parameter table in parameter database
        """
        raster = self.get_raster(Names.manningName)
        if raster is None:
            raster = RasterExtension.reclassify(self.mapped_landuse_raster, self.parameter.get_parameter_lookup("MANNING","landuse"), self.subbasin_raster)
            raster = self.save_raster(raster, Names.manningName, True, True)

        return raster
    
    @property
    def velocity_raster(self)->Raster:
        """
        velocity, need verify
        """
        raster = self.get_raster(Names.velocityName)

        if raster is None:
            min = 0.005
            max = 3
            velocity = self.slope_radius_raster.sqrt() * self.reach_depth_raster**0.6667 / self.manning_raster
            velocity = self.subbasin_raster.con("value > 0", velocity, velocity.configs.nodata)
            raster = velocity.min(max).max(min)
            raster = self.save_raster(raster, Names.velocityName, True, True)
        return raster
           
    @property
    def wetness_index_raster(self)->Raster:
        """
        wetness index
        """
        raster = self.get_raster(Names.wetnessIndexName)

        if raster is None:
            raster = (self.flow_acc_raster * self.inputs.cellsize_m2 / self.slope_radius_raster).ln()
            raster = self.subbasin_raster.con("value > 0", raster, raster.configs.nodata)
            raster = self.save_raster(raster, Names.wetnessIndexName, True, True)

        return raster
    
    @property
    def initial_soil_moisture_raster(self)->Raster:
        """
        initial soil moisture interpolated based on wetness index
        """

        raster = self.get_raster(Names.moistureInitialName)

        if raster is None:
            minSaturation = 0.05 
            maxSaturation = 1

            minWetnessIndex = self.wetness_index_raster.configs.minimum
            maxWetnessIndex = self.wetness_index_raster.configs.maximum * 0.8

            wti = self.wetness_index_raster.min(maxWetnessIndex)
            ratio = (wti - minWetnessIndex) * (maxSaturation - minSaturation) / (maxWetnessIndex - minWetnessIndex) + minSaturation
            raster = ratio * self.field_capacity_raster
            raster = self.subbasin_raster.con(f"value == {self.subbasin_raster.configs.nodata}", raster.configs.nodata, raster)

            raster = self.save_raster(raster, Names.moistureInitialName, True, True)
        return raster
    
    @property
    def potential_ruoff_coefficient_raster(self)->Raster:
        """
        potential runoff coefficient
        """

        raster = self.get_raster(Names.PRCName)
        subbasin_nodata = self.subbasin_raster.configs.nodata

        if raster is None:
            slope_raster = self.slope_degree_raster
            soil_raster = self.mapped_soil_raster
            landuse_raster = self.mapped_landuse_raster
            flow_dir_raster = self.flow_direction_raster
            reach_raster = self.reach_raster
            reach_width_raster = self.reach_width_raster
            raster = self.inputs.create_new_raster()

            for row in range(self.inputs.rows):
                for col in range(self.inputs.columns):
                    if self.subbasin_raster[row, col] == subbasin_nodata:
                        continue
                    slope = slope_raster[row, col]
                    landuse = landuse_raster[row, col]
                    soil = soil_raster[row,col]
                    reach = reach_raster[row,col]
                    reach_width = reach_width_raster[row,col]
                    flow_dir = flow_dir_raster[row,col]

                    reachWaterSurfaceFraction = self.__calculate_reach_water_surface_fraction(reach, reach_width,flow_dir)
                    raster[row,col] = self.parameter.get_potential_runoff_coefficient(landuse, soil, slope, reachWaterSurfaceFraction)
           
            raster = self.save_raster(raster, Names.PRCName, True, True)

        return raster
    
    @property
    def potential_ruoff_coefficient_accumulate_average_raster(self)->Raster:
        """
        accumulative potential runoff coefficient
        """

        raster = self.get_raster(Names.PRCAccAvgName)

        if raster is None:
            raster = self.__value_accumulate_d8(self.potential_ruoff_coefficient_raster)

            raster = self.save_raster(raster, Names.PRCAccAvgName, True, True)
        return raster
    
    @property
    def depression_storage_capacity_raster(self)->Raster:
        """
        depression storage
        """
        raster = self.get_raster(Names.DSCName)

        subbasin_nodata = self.subbasin_raster.configs.nodata

        if raster is None:
            slope_raster = self.slope_degree_raster
            soil_raster = self.mapped_soil_raster
            landuse_raster = self.mapped_landuse_raster
            flow_dir_raster = self.flow_direction_raster
            reach_raster = self.reach_raster
            reach_width_raster = self.reach_width_raster
            raster = self.inputs.create_new_raster()

            for row in range(self.inputs.rows):
                for col in range(self.inputs.columns):
                    if self.subbasin_raster[row, col] == subbasin_nodata:
                        continue
                    slope = slope_raster[row, col]
                    landuse = landuse_raster[row, col]
                    soil = soil_raster[row,col]
                    reach = reach_raster[row,col]
                    reach_width = reach_width_raster[row,col]
                    flow_dir = flow_dir_raster[row,col]

                    reachWaterSurfaceFraction = self.__calculate_reach_water_surface_fraction(reach, reach_width,flow_dir)
                    raster[row,col] = self.parameter.get_depression_storage_capacity(
                        landuse, soil, slope, reachWaterSurfaceFraction)
           
            raster = self.save_raster(raster, Names.DSCName, True, True)

        return raster
    
    @property
    def depression_storage_capacity_accumulate_average_raster(self)->Raster:
        """
        depresion storage accumulative
        """

        raster = self.get_raster(Names.DSCAccAvgName)

        if raster is None:
            raster = self.__value_accumulate_d8(self.depression_storage_capacity_raster)

            raster = self.save_raster(raster, Names.DSCAccAvgName, True, True)
        return raster
    
    @property
    def cn2_raster(self)->Raster:
        """
        CN2 parameter
        """

        raster = self.get_raster(Names.cn2Name)
        subbasin_nodata = self.subbasin_raster.configs.nodata
        if raster is None:
            slope_raster = self.slope_degree_raster
            soil_raster = self.mapped_soil_raster
            landuse_raster = self.mapped_landuse_raster
            flow_dir_raster = self.flow_direction_raster
            reach_raster = self.reach_raster
            reach_width_raster = self.reach_width_raster
            raster = self.inputs.create_new_raster()

            for row in range(self.inputs.rows):
                for col in range(self.inputs.columns):
                    slope = slope_raster[row, col]
                    landuse = landuse_raster[row, col]
                    soil = soil_raster[row,col]
                    reach = reach_raster[row,col]
                    reach_width = reach_width_raster[row,col]
                    flow_dir = flow_dir_raster[row,col]
                    if self.subbasin_raster[row, col] == subbasin_nodata:
                        continue

                    reachWaterSurfaceFraction = self.__calculate_reach_water_surface_fraction(reach, reach_width,flow_dir)
                    cn2 = self.parameter.get_cn2(landuse, soil)
                    if reachWaterSurfaceFraction > 0:  # stream
                        cn2 = reachWaterSurfaceFraction * 100.0 + (1 - reachWaterSurfaceFraction) * cn2
                        cn2 = max(cn2, 35.0)
                        cn2 = min(cn2, 100.0)

                    raster[row,col] = cn2
           
            raster = self.save_raster(raster, Names.cn2Name, True, True)

        return raster
    
    @property
    def cn2_accumulate_average_raster(self)->Raster:
        raster = self.get_raster(Names.Cn2AccAvgName)

        if raster is None:
            raster = self.__value_accumulate_d8(self.cn2_raster)

            raster = self.save_raster(raster, Names.Cn2AccAvgName, True, True)
        return raster
    
    def __calculate_reach_water_surface_fraction(self, reach, reach_width, flow_dir):
        reachWaterSurfaceFraction = 0.0
        if reach > 0:
            if flow_dir in [1, 4, 16, 64]:  # corner direction
                reachWaterSurfaceFraction = reach_width / self.inputs.cell_size
                reachWaterSurfaceFraction = 1.41421356 * reachWaterSurfaceFraction - 0.5 * reachWaterSurfaceFraction * reachWaterSurfaceFraction
            else:
                reachWaterSurfaceFraction = reach_width / self.inputs.cell_size
        
            reachWaterSurfaceFraction = max(0.0, reachWaterSurfaceFraction)
            reachWaterSurfaceFraction = min(1.0, reachWaterSurfaceFraction)     
        return reachWaterSurfaceFraction      

    def __value_accumulate_d8(self, input_raster:Raster, is_average:bool = True)->Raster:
        row, col, x, y = 0, 0, 0, 0
        z, z2 = 0.0, 0.0
        v, v2 = 0.0, 0.0
        i = 0
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        inflowingVals = [16, 32, 64, 128, 1, 2, 4, 8]
        numInNeighbours = 0.0
        isAvg = False
        flag = False
        flowDir = 0.0

        isAvg = is_average

        try:
            pntr = self.flow_direction_raster
            valueRas = input_raster

            noData = pntr.configs.nodata
            rows = pntr.configs.rows
            cols = pntr.configs.columns
    
            output = self.inputs.create_new_raster
            temp = np.full((rows, cols), 1.0)

            tmpGrid1 = self.inputs.create_new_raster

            for row in range(rows):
                for col in range(cols):
                    if pntr[row, col] != noData and valueRas[row, col] != noData:
                        z = 0
                        for i in range(8):
                            if pntr[row + dY[i], col + dX[i]] == inflowingVals[i]:
                                z += 1
                        tmpGrid1[row, col] = z
                        output[row, col] = valueRas[row, col]
                    else:
                        temp[row, col] = noData

            for row in range(rows):
                for col in range(cols):
                    if tmpGrid1[row, col] == 0:
                        tmpGrid1[row, col] = -1
                        x, y = col, row
                        while True:
                            z = temp[y, x]
                            v = output[y, x]
                            flowDir = pntr[y, x]
                            if flowDir > 0:
                                i = int(math.log(flowDir) / math.log(2))
                                x += dX[i]
                                y += dY[i]
                                z2 = temp[y, x]
                                temp[y, x] = z2 + z
                                v2 = output[y, x]
                                output[y, x] = v2 + v
                                numInNeighbours = tmpGrid1[y, x] - 1
                                tmpGrid1[y, x] = numInNeighbours
                                if numInNeighbours == 0:
                                    tmpGrid1[y, x] = -1
                                    flag = True
                                else:
                                    flag = False
                            else:
                                flag = False
                            if not flag:
                                break

            if isAvg:
                for row in range(rows):
                    for col in range(cols):
                        flowDir = output[row, col]
                        if flowDir != noData and valueRas[row, col] != noData:
                            output[row, col] = output[row, col] / temp[row, col]
                        else:
                            output[row, col] = noData

            return output

        except MemoryError:
            print("memory error")
        except Exception as e:
            print(e)
        finally:
            pass
    
#endregion

#region delineation

    @property
    def flow_direction_no_changed_raster(self)->Raster:
        """Flow direction based on the filled dem. No further modifications are applied."""
        flow_dir_raster = self.get_raster(Names.flowDirD8NoChangeName)     

        if flow_dir_raster is None:
            flow_dir_raster = self.wbe.d8_pointer(dem = self.dem_clipped_burned_filled_raster)
            flow_dir_raster = self.save_raster(flow_dir_raster,Names.flowDirD8NoChangeName, True, True)

        return flow_dir_raster

    @property
    def flow_direction_raster(self)->Raster:
        """Flow direction modified by structures"""
        flow_dir_raster = self.get_raster(Names.flowDirD8Name)     

        if flow_dir_raster is None:
            if self.structure_combined is not None:                
                #if there are some structure, use the structure to get the flow direction
                #merge all structure boundary and outlets to get the final flow direction.
                flow_dir_raster = self.flow_direction_no_changed_raster
                flow_dir_raster = self.structure_combined.generate_flow_direction_raster_shawn()
            else:
                #if there are no structure, use the dem directly
                logger.info("Creating flow direction based on dem ...")
                flow_dir_raster = self.wbe.d8_pointer(dem = self.dem_clipped_burned_filled_raster)
            flow_dir_raster = self.save_raster(flow_dir_raster, Names.flowDirD8Name, True, True)
        
        return flow_dir_raster

    @property
    def flow_acc_raster(self)->Raster:
        """Flow accumulation raster"""
        raster = self.get_raster(Names.flowAccName)

        if raster is None:  
            logger.info("Creating flow accumulation raster ...")
            raster = self.wbe.d8_flow_accum(raster = self.flow_direction_raster, 
                                            input_is_pointer = True,
                                            out_type='cells')
            raster = self.save_raster(raster, Names.flowAccName, True, True)

        return raster
    
    @property
    def flow_length_raster(self)->Raster:        
        raster = self.get_raster(Names.flowLengthName)

        if raster is None:
            logger.info("Creating flow length raster ...")
            raster = Delineation.get_flow_path_length(self.flow_direction_no_changed_raster, self.dem_clipped_raster_for_model)
            raster = self.save_raster(raster, Names.flowLengthName, True, True)

        return raster    
    
    @property
    def stream_network_user_raster(self)->Raster:
        """stream network provided by user for burn-in"""
        raster = self.get_raster(Names.streamNetworUserRasName)        
        if raster is None:
            raster = self.wbe.vector_lines_to_raster(
                input = self.inputs.stream_network_user_vector, 
                base_raster = self.inputs.dem_raster)
            #raster = self.wbe.buffer_raster(raster, self.inputs.cell_size)
            raster = self.save_raster(raster, Names.streamNetworUserRasName, True, True)
        return raster
    
    @property
    def stream_main_raster(self)->Raster:
        """stream network delineated with fixed 200ha threshold to define the pour points."""

        raster = self.get_raster(Names.streamMainRasName)
   
        if raster is None:       
            #get streams from the flow accumuation 
            logger.info(f"Extracting stream raster with threashold area of {self.main_stream_threshold_area_ha}ha ...")    
            raster = self.wbe.extract_streams(
                flow_accumulation = self.flow_acc_raster, 
                threshold = self.main_stream_threshold_area_ha / self.inputs.cellsize_ha)
                 
            #save it
            raster = self.save_raster(raster, Names.streamMainRasName, True, True)

        return raster   

    @property
    def stream_network_raster(self)->Raster:
        """stream network delineated with user defined threshold. The structures are also considered."""

        raster = self.get_raster(Names.streamNetworkRasName)
   
        if raster is None:       
            #get streams from the flow accumuation 
            logger.info(f"Extracting stream raster with threashold area of {self.stream_threshold_area_ha}ha ...")    
            raster = self.wbe.extract_streams(
                flow_accumulation = self.flow_acc_raster, 
                threshold = self.stream_threshold_area_ha / self.inputs.cellsize_ha)
            
            #fix it as some of the structure streams may be below the threshold 
            logger.info("Fixing stream network so the structure stream segment that is below threashold will be added back ...")
            raster = Delineation.build_stream_network_link_to_outlets(raster, self.flow_direction_raster,self.structure_combined_outlet_raster)

            #save it
            raster = self.save_raster(raster, Names.streamNetworkRasName, True, True)

        return raster    
    
    @property
    def stream_network_vector(self)->Vector:
        vector = self.get_vector(Names.streamNetworkShpName)
   
        if vector is None:            
            vector = self.wbe.raster_streams_to_vector(streams = self.stream_network_raster, d8_pointer = self.flow_direction_raster)
            self.save_vector(vector, Names.streamNetworkRasName)

        return vector    
    
    @property
    def stream_pour_point_raster(self)->Raster:
        """
        the raster of all stream outlets
        """
        raster = self.get_raster(Names.streamPourPointRasName)

        if raster is None:
            logger.info("Converting stream outlets pour point vector to raster ...")
            raster = self.wbe.vector_points_to_raster(input = self.stream_pour_points_vector,
                                                      base_raster = self.inputs.dem_raster)
            raster = self.save_raster(raster, Names.streamPourPointRasName, True, True)

        return raster

    @property
    def stream_pour_points_vector(self)->Vector:
        """
        the stream outlets considering structures, user-defined outlets and reservoirs
        """
        vector = self.get_vector(Names.streamPourPointShpName)

        if vector is None:
            logger.info("Merging all possible stream outlets pour points to a single vector ...")

            #merge the pour points and user-provided reservoir and outlets
            outlet_vectors = []

            #add stream outlets delineated using main stream
            logger.info("Creating pour points with main stream (200ha threshold) ...")
            pour_points_vector = Delineation.get_pour_points(
                self.stream_main_raster, 
                self.flow_direction_raster, 
                [structure.boundary_raster for structure in self.structures.values()])
            outlet_vectors.append(pour_points_vector)
            
            #add user-defined watershed outlets and snap it to the stream network
            #this is basically is the one user selected on the map
            #this should be the outlet of the whole watershed.
            #we will check if the shapefile only has one point.
            # if self.inputs.outlet_vector is not None:
            #     outlet_vectors.append(self.wbe.jenson_snap_pour_points(
            #         pour_pts = self.inputs.outlet_vector,
            #         streams = self.stream_network_raster,
            #         snap_dist = 2000))

            #add reach bmp points but exclude wetland
            for reach_bmp in self.inputs.reach_bmp_vectors:
                if reach_bmp is not None:
                    outlet_vectors.append(self.wbe.jenson_snap_pour_points(
                        pour_pts = reach_bmp,
                        streams = self.stream_network_raster,
                        snap_dist = 2000))

            #add structure outlets
            #there is no need to snap as the outlet has been processed with the network
            if len(self.structures) > 0:
                outlet_vectors.extend([structure.outlet_vector for structure in self.structures.values()])

            vector = VectorExtension.merge_vectors(outlet_vectors)
            self.save_vector(vector, Names.streamPourPointShpName)

        return vector

    @property
    def subbasin_vector(self)->Vector:
        vector = self.get_vector(Names.subbasinShpName)

        if vector is None:
            vector = RasterExtension.raster_to_vector(self.subbasin_raster)
            self.save_vector(vector, Names.subbasinShpName)

        return vector

    @property
    def subbasin_raster(self)->Raster:
        raster = self.get_raster(Names.subbasinRasName)

        if raster is None:
            #get subbasins
            logger.info("creating subbasins ...")
            raster = self.wbe.watershed(
                d8_pointer=self.flow_direction_raster, 
                pour_points=self.stream_pour_points_vector)

            #reorder the subbasin starting from 1
            logger.info("Reordering subbasin ...")
            Delineation.reorder_raster_id(raster)

            #fix the subbasins considering the structures
            logger.info("Fixing subbasin for structure and assign subbasin id to sturcture ...")
            for type,structure in self.structures.items():
                #logger.info(type)
                structure.repair_subbasin_and_assign_subbasin_id_contribution_area_to_sturcture(raster, self.flow_acc_raster)  

            #save the subbasin raster
            raster = self.save_raster(raster, Names.subbasinRasName, True, True)

        return raster
    
    @property
    def slope_percent_raster(self)->Raster:
        raster = self.get_raster(Names.slopePercentName)

        if raster is None:
            logger.info("creating slope (percent) raster ...")
            raster = self.slope_radius_raster * 100
            raster = self.save_raster(raster, Names.slopePercentName, True, True)

        return raster
    
    @property
    def slope_radius_raster(self)->Raster:
        raster = self.get_raster(Names.slopeRadiusName)

        if raster is None:
            logger.info("creating slope (radius) raster ...")
            raster = (self.slope_degree_raster * math.pi / 180).tan()
            raster = self.save_raster(raster, Names.slopeRadiusName, True, True)

        return raster
    
    @property
    def slope_degree_raster(self)->Raster:        
        raster = self.get_raster(Names.slopeDegName)

        if raster is None:
            logger.info("creating slope (degree) raster ...")
            raster = self.wbe.slope(dem = self.dem_clipped_burned_filled_raster)
            raster = self.save_raster(raster, Names.slopeDegName, True, True)

        return raster

    @property
    def stream_order_raster(self)->Raster:
        raster = self.get_raster(Names.streamOrderRasName)  

        if raster is None:
            logger.info("creating stream order raster ...")
            raster = self.wbe.strahler_stream_order(d8_pntr = self.flow_direction_raster, streams_raster = self.stream_network_raster)
            raster = self.save_raster(raster, Names.streamOrderRasName, True, True)

        return raster
            

#endregion

#region reach

    @property
    def reach_vector(self)->Vector:
        vector = self.get_vector(Names.reachShpName) 

        if vector is None:
            logger.info("creating reach vector ...")
            vector = self.wbe.raster_streams_to_vector(streams = self.reach_raster, d8_pointer = self.flow_direction_raster)
            self.save_vector(vector, Names.reachShpName)

        return vector

    @property
    def reach_raster(self)->Raster:
        raster = self.get_raster(Names.reachRasName) 

        if raster is None:
            logger.info("creating reach raster ...")
            raster = self.stream_network_raster.con("value >= 0", self.subbasin_raster, self.subbasin_raster.configs.nodata)
            raster = self.save_raster(raster, Names.reachRasName, True, True)

        return raster

    @property
    def reach_width_raster(self)->Raster:
        """
        calculate reach width for 2-year design storm, will use-defined parameter later
        """
        return self._get_reach_width_depth_raster(False)
    
    @property
    def reach_depth_raster(self)->Raster:
        """
        calculate reach depth for 2-year design storm, will use-defined parameter later
        """
        return self._get_reach_width_depth_raster(True)
    
    def _get_reach_width_depth_raster(self, isdepth:bool)->Raster:
        """
        calculate reach width/depth raster using given design storm

        replace StreamWidth and StreamDepth
        """
        file_name = Names.reachWidthName
        if isdepth:
            file_name = Names.reachDepthName

        raster = self.get_raster(file_name)
        if raster is None:
            
            logger.info(f"creating reach {"depth" if isdepth else "width"} raster ...")
            parameter = self.parameter.get_reach_width_parameter(self.design_storm_return_period)          
            if isdepth:
                parameter = self.parameter.get_reach_depth_parameter(self.design_storm_return_period) 
            raster = (self.flow_acc_raster * self.inputs.cellsize_km2 * parameter.A) ** parameter.B
            raster = self.stream_network_raster.con("value > 0", raster, raster.configs.nodata)
            raster = self.save_raster(raster, file_name, True, True)

        return raster

    @property
    def reach_parameter_df(self)->pd.DataFrame:
        df = self.get_df(Names.reachParameterCsvName)

        if df is None:            
            df = Reach(self.dem_clipped_raster, 
                       self.subbasin_raster,
                       self.stream_network_raster,
                       self.stream_order_raster,
                       self.flow_length_raster,
                       self.flow_direction_raster,
                       self.flow_acc_raster,
                       self.reach_width_raster,
                       self.reach_depth_raster,
                       self.velocity_raster
                       ).reach_parameter_df 
            self.save_df(df, Names.reachParameterCsvName)

        return df  

#endregion

#region bmps

    @property
    def marginal_crop_land_raster(self):
        """
        generate marginal crop land

        replace MarginalCropland
        """
        raster = self.get_raster(Names.marCroplandRasName)  

        if raster is None and self.marginal_crop_land_landuse_ids is not None and len(self.marginal_crop_land_landuse_ids) > 0:
            filtered_landuse_raster = RasterExtension.filter_by_values(self.mapped_landuse_raster, self.marginal_crop_land_landuse_ids)
            buffered_landuser_raster = self.wbe.buffer_raster(input = filtered_landuse_raster, buffer_size = self.marginal_crop_land_buffer_size_m) #need test

            slope_threshod_deg = max(math.atan(self.marginal_crop_land_slope_threshod / 100) * 180.0 / math.pi, 0.0)
            
            raster = self.field_without_agriculture_raster.deep_copy()
            raster = buffered_landuser_raster.con("value > 0", raster, raster.configs.nodata)
            raster = self.slope_degree_raster.con(f"value >= {slope_threshod_deg}", raster, raster.configs.nodata)    
            raster = self.mask_refined_with_subbasin_raster.con(f"value > 0", raster, raster.configs.nodata)

            raster = self.save_raster(raster, Names.marCroplandRasName, True, True)
                    
        return raster
    
    @property
    def pasture_land_raster(self):
        """
        generate pasture land

        replace BuildMapOnLandUse
        """
        raster = self.get_raster(Names.pastureLandRasName)  

        if raster is None and self.pasture_land_landuse_ids is not None and len(self.pasture_land_landuse_ids) > 0:
            filtered_landuse_raster = RasterExtension.filter_by_values(self.mapped_landuse_raster, self.pasture_land_landuse_ids)
            
            raster = self.field_without_agriculture_raster.deep_copy()
            raster = filtered_landuse_raster.con("value > 0", raster, raster.configs.nodata)
            raster = self.mask_refined_with_subbasin_raster.con(f"value > 0", raster, raster.configs.nodata)

            raster = self.save_raster(raster, Names.pastureLandRasName, True, True)
                    
        return raster

#region Riparian Buffer

    @property
    def riparian_buffer_raster(self)->Raster:
        raster = self.get_raster(Names.riparianBufferRasterName)
    
        if raster is None:
            if self.inputs.raparian_buffer_vector is not None:
                exist, id_field_name = VectorExtension.check_id(self.inputs.raparian_buffer_vector)
                if not exist:
                    raise ValueError(f"Couldn't find ID column in {self.inputs.raparian_buffer_vector.file_name} ...")
                raster = self.wbe.vector_polygons_to_raster(input = self.inputs.raparian_buffer_vector, 
                                                            field_name = id_field_name,
                                                            base_raster = self.inputs.dem_raster)
                raster = self.mask_refined_with_subbasin_raster.con(f"value == {self.inputs.nodata}", self.inputs.nodata, raster)
                raster = self.save_raster(raster, Names.riparianBufferRasterName, True, True)

        return raster
    
    @property
    def riparian_buffer_drainage_area_raster(self)->Raster:
        raster = self.get_raster(Names.riparianBufferDrainageRasterName)

        if raster is None:
            self.__create_riparian_buffer_distribution_paramter()
            raster = self.get_raster(Names.riparianBufferDrainageRasterName)

        return raster

    @property
    def riparian_buffer_parts_raster(self)->Raster:
        raster = self.get_raster(Names.riparianBufferPartRasterName)

        if raster is None:
            self.__create_riparian_buffer_distribution_paramter()
            raster = self.get_raster(Names.riparianBufferPartRasterName)

        return raster   
    
    @property
    def riparian_buffer_parameter_df(self)->pd.DataFrame:
        df = self.get_df(Names.riparianBufferParameterCSVName)

        if df is None:
            self.__create_riparian_buffer_distribution_paramter()
            df = self.get_df(Names.riparianBufferParameterCSVName)

        return df  

    def __create_riparian_buffer_distribution_paramter(self, width = 10):
        """
        create distribution raster and parameter table for riparian buffer. 
        
        It finds all parts of the buffer in each subbasin and corresponding drainage area. And then summarize the paramters. 

        Replace Plugin VFSandRBS
        """

        row, col, x, y = 0, 0, 0, 0
        z = 0.0
        i, c = 0, 0
        flag = False
        flowDir = 0.0

        if self.riparian_buffer_raster is None:
            return

        logger.info("Creating riparian buffer distribution and parameters ... ")

        rows = self.flow_direction_raster.configs.rows
        cols = self.flow_direction_raster.configs.columns
        riparian_buffer_nodata = self.riparian_buffer_raster.configs.nodata
        subbasin_nodata = self.subbasin_raster.configs.nodata

        riparian_buffer_drainage_area_raster = self.wbe.new_raster(self.subbasin_raster.configs)
        riparian_buffer_parts_raster = self.wbe.new_raster(self.subbasin_raster.configs)

        #flag if the cells has been traced
        flag2D = np.zeros((rows, cols), dtype=bool)


        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        LnOf2 = 0.693147180559945
        path = []
        pathInside = []
        vfsOrRbsDrain = {}
        vfsOrRbsInside = {}
        vfsOrRbsLength = {}
        celldist = (self.riparian_buffer_raster.configs.resolution_x + self.riparian_buffer_raster.configs.resolution_y) / 2.0

        for row in range(rows):
            for col in range(cols):
                if self.subbasin_raster[row, col] != subbasin_nodata and not flag2D[row, col]:
                    path = []
                    pathInside = []

                    x = col
                    y = row

                    path.append((y, x))

                    flag = True
                    length = 0.0

                    while flag:
                        flowDir = self.flow_direction_raster[y, x]
                        if flowDir > 0:
                            c = int(math.log(flowDir) / LnOf2)

                            lastVFSorRBSID = self.riparian_buffer_raster[y, x]
                            isVisited = flag2D[y, x]

                            if lastVFSorRBSID != riparian_buffer_nodata and not isVisited:
                                pathInside.append((y, x))
                                length += celldist * (1.41421356 if c in [1, 4, 16, 64] else 1)

                            #get x,y for downstream cell
                            x += dX[c]
                            y += dY[c]

                            #check to see if the next cell is edge, different riparian buffer. If yes, add it to the list
                            if (self.riparian_buffer_raster[y, x] == riparian_buffer_nodata 
                                or self.riparian_buffer_raster[y, x] != lastVFSorRBSID 
                                or self.flow_direction_raster[y, x] == 0) and lastVFSorRBSID != riparian_buffer_nodata:
                                #recalculate the current position
                                loc = (y - dY[c], x - dX[c])

                                if loc in vfsOrRbsDrain:
                                    vfsOrRbsDrain[loc].extend(path)
                                else:
                                    vfsOrRbsDrain[loc] = path.copy()

                                if loc in vfsOrRbsInside:
                                    vfsOrRbsInside[loc].extend(pathInside)
                                else:
                                    vfsOrRbsInside[loc] = pathInside.copy()

                                if loc in vfsOrRbsLength:
                                    vfsOrRbsLength[loc] = max(length, vfsOrRbsLength[loc])
                                else:
                                    vfsOrRbsLength[loc] = length
                                break

                            if not isVisited:
                                path.append((y, x))
                        else:
                            flag = False

                    #update the flag
                    for loc in path:
                        flag2D[loc[0], loc[1]] = True

        #create the resulting parameter dictionary
        riparain_buffer_parameter_df = pd.DataFrame(columns=['ID','RIBUF_ID','Subbasin','Area_ha','Drainage_Area','Area_Ratio','Length'])
        riparain_buffer_parameter_df.set_index('ID', inplace=True)        

        cellArea = self.riparian_buffer_raster.configs.resolution_x * self.riparian_buffer_raster.configs.resolution_y
        vfsOrRbsPartID = 0
        for outLoc in vfsOrRbsDrain.keys():
            vfsOrRbsPartID += 1

            vfsOrRbsID = int(self.riparian_buffer_raster[outLoc[0], outLoc[1]])
            subID = int(self.subbasin_raster[outLoc[0], outLoc[1]])
            drainageArea = len(vfsOrRbsDrain[outLoc]) * cellArea / 10000
            length = vfsOrRbsLength[outLoc]

            riparain_buffer_parameter_df.loc[vfsOrRbsPartID,'RIBUF_ID'] = vfsOrRbsID
            riparain_buffer_parameter_df.loc[vfsOrRbsPartID,'Subbasin'] = subID
            riparain_buffer_parameter_df.loc[vfsOrRbsPartID,'Drainage_Area'] = drainageArea
            riparain_buffer_parameter_df.loc[vfsOrRbsPartID,'Length'] = length

            for loc in vfsOrRbsDrain[outLoc]:
                riparian_buffer_drainage_area_raster[loc[0], loc[1]] = vfsOrRbsPartID

            for loc in vfsOrRbsInside[outLoc]:
                riparian_buffer_parts_raster[loc[0], loc[1]] = vfsOrRbsPartID

        riparain_buffer_parameter_df["Width"] = width
        riparain_buffer_parameter_df["Area_ha"] = width * riparain_buffer_parameter_df['Length'] * 0.0001
        riparain_buffer_parameter_df["Area_Ratio"] = riparain_buffer_parameter_df["Drainage_Area"] / riparain_buffer_parameter_df["Area_ha"]
        riparain_buffer_parameter_df["VegetationID"] = 6

        #get slope, soil_k, soil_porosity and root_depth using the zone statistics
        slope_df = RasterExtension.get_zonal_statistics(self.slope_degree_raster, riparian_buffer_parts_raster, "mean","Slope")
        k_df = RasterExtension.get_zonal_statistics(self.soil_k_raster, riparian_buffer_parts_raster, "mean","Sol_K")
        prosity_df = RasterExtension.get_zonal_statistics(self.soil_porosity_raster, riparian_buffer_parts_raster, "mean","Sol_porosity")
        root_depth_df = RasterExtension.get_zonal_statistics(self.landuse_rootdepth_raster, riparian_buffer_parts_raster, "mean","Root_Depth")

        riparain_buffer_parameter_df = riparain_buffer_parameter_df.merge(slope_df, left_index=True, right_index=True, how="inner")
        riparain_buffer_parameter_df = riparain_buffer_parameter_df.merge(k_df, left_index=True, right_index=True, how="inner")
        riparain_buffer_parameter_df = riparain_buffer_parameter_df.merge(prosity_df, left_index=True, right_index=True, how="inner")
        riparain_buffer_parameter_df = riparain_buffer_parameter_df.merge(root_depth_df, left_index=True, right_index=True, how="inner")

        #add other columns with default values
        riparain_buffer_parameter_df["Scenario"] = DefaultScenarioId
        riparain_buffer_parameter_df["Year"] = 0

        #adjust the sequence of the columns
        riparain_buffer_parameter_df.reset_index(inplace=True)
        riparain_buffer_parameter_df = riparain_buffer_parameter_df[['Scenario','ID','Year','RIBUF_ID','Subbasin','VegetationID','Length','Width','Area_ha','Drainage_Area','Area_Ratio','Slope','Sol_K','Sol_porosity','Root_Depth']]

        #save to bmp database
        self.save_raster(riparian_buffer_drainage_area_raster, Names.riparianBufferDrainageRasterName)
        self.save_raster(riparian_buffer_parts_raster, Names.riparianBufferPartRasterName)
        self.save_df(riparain_buffer_parameter_df, Names.riparianBufferParameterCSVName)

#endregion

#endregion

#region interpolation

#endregion

    @property
    def offsite_watering_raster(self)->Raster:
        raster = self.get_raster(Names.offsiteWinteringRasName)

        if raster is None:
            raster = VectorExtension.vector_to_raster(self.inputs.offsite_watering_vector, self.dem_clipped_raster_for_model)
            self.save_raster(raster, Names.offsiteWinteringRasName)
        return raster

    def delineate_watershed(self):
        """watershed delineation which basically delineate stream and subbasins"""
        subbasin = self.subbasin_vector
        dem_for_model = self.dem_clipped_raster_for_model
        reach = self.reach_vector
        soil_moisture = self.initial_soil_moisture_raster
        slope_percent = self.slope_percent_raster
        wetland = self.wetland_raster
        usle_p = self.uslep_raster
        offsite_watering = self.offsite_watering_raster
    







from ...bmp.bmp_type import BMPType, DefaultScenarioId

class BMP_scenarios:
    def __init__(self, bmp_type:BMPType, distribution:str, parameter:str):
        self.ID = DefaultScenarioId
        self.NAME = "scenario"
        self.BMP = bmp_type.value
        self.DISTRIBUTION = distribution
        self.PARAMETER = parameter

        


from ...bmp.bmp_type import DefaultScenarioId

class BMPManagementBase:
    def __init__(self):
        """Base class for all bmp management tables. They start with same columns."""
        self.Scenario = DefaultScenarioId
        self.Location = -1 

class BMPManagementBaseWithYear(BMPManagementBase):
    """Base class for all bmp management tables. They start with same columns."""
    def __init__(self):
        super().__init__()
        self.Year = 1


import os
import logging
from .interpolation import Interpolation
from whitebox_workflows import Raster
from ..raster_extension import RasterExtension

class AverageUniform(Interpolation):
    def __init__(self, weight_file: str) -> None:
        super().__init__(weight_file)

    def write_weight_file(self, mask_raster:Raster, station_coordinates:list):
        rows = mask_raster.configs.rows
        cols = mask_raster.configs.columns
        numShapes = len(station_coordinates)
        rowCount = RasterExtension.get_number_of_valid_cell(mask_raster)

        sb = []
        sb.append(str(rowCount))
        sb.append(str(numShapes))

        if numShapes != 0:
            weight = 10000 // numShapes / 10000
            weightf = 1 - (weight * (numShapes - 1)) if weight * numShapes != 1 else weight

            for row in range(rows):
                for col in range(cols):
                    if mask_raster[row, col] > 0:
                        finalValue = str(weight)
                        if numShapes > 1:
                            finalValue += "\t" + "\t".join([str(weight)] * (numShapes - 1)) + "\t" + str(weightf)
                        sb.append(finalValue)

        with open(self.weight_file, 'w') as out:
            out.write("\n".join(sb))


import pandas as pd
import os

class FileIn:
    def __init__(self, 
                 folder:str,
                 model_type:str,
                 cell_size:int, 
                 cell_num:int, 
                 subarea_num:int, 
                 subbasin_num:int, 
                 start_date:pd.Timestamp, 
                 end_date:pd.Timestamp,
                 data_type_station_ids:dict,
                 interval:str = "Daily") -> None:
        self.folder = folder
        self.model_type = model_type
        self.cell_size = cell_size
        self.cell_num = cell_num
        self.subarea_num = subarea_num
        self.subbasin_num = subbasin_num
        self.start_date = start_date
        self.end_date = end_date
        self.interval = interval
        self.data_type_station_ids = data_type_station_ids
        

    def write_file(self):
        with open(os.path.join(self.folder, "file.in"),'w') as f:
            if self.model_type == "subarea":
                f.writelines(f"VERSIONTYPE|1")
            else:
                f.writelines(f"VERSIONTYPE|0")

            f.writelines(f"\nISPREPAREFORSUBAREAINPUT|0")
            f.writelines(f"\nCELLSIZE|{self.cell_size}")
            f.writelines(f"\nCELLNUMBER|{self.cell_num}")
            f.writelines(f"\nSUBAREANUMBER|{self.subarea_num}")
            f.writelines(f"\nSUBBASINCOUNT|{self.subbasin_num}")
            f.writelines(f"\nINTERVAL|{self.interval}")
            f.writelines(f"\nSTARTTIME|{self.start_date.strftime("%Y/%m/%d")}")
            f.writelines(f"\nENDTIME|{self.end_date.strftime("%Y/%m/%d")}")

            for data_type, ids in self.data_type_station_ids.items():
                f.writelines(f"\nSITECOUNT|{len(ids)}|{data_type}")
                for id in ids:
                    f.writelines(f"\nSITENAME|{id}|{data_type}")

            




from sqlalchemy import Column, Integer, REAL, TEXT
from .hydroclimate_table import HydroClimateTable

class Stations(HydroClimateTable):
    __tablename__ = 'STATIONS'
    ID = Column(Integer, primary_key=True)
    NAME = Column(TEXT)
    XPR = Column(REAL)
    YPR = Column(REAL)
    LAT = Column(REAL)
    LONG = Column(REAL)
    ELEVATION = Column(REAL)
    AREA = Column(REAL)


from .folder_base import FolderBase
from .names import Names
from .raster_extension import RasterExtension
from .vector_extension import VectorExtension
from whitebox_workflows import Raster, Vector
from .lookup import Lookup
import logging
import numpy as np
from .bmp.bmp_type import BMPType
from .bmp.bmp_reach_reservoir import ReachBMPReservoir

logger = logging.getLogger(__name__)

class Inputs(FolderBase):
    """
    Info from input folder, these are the original input file with standard name
    """
    def __init__(self, input_folder:str) -> None:
        super().__init__(input_folder)

        self.__validate()

        self.cell_size = self.dem_raster.configs.resolution_x        
        self.cellsize_m2 = self.dem_raster.configs.resolution_x * self.dem_raster.configs.resolution_y
        self.cellsize_km2 = self.cellsize_m2 / 1e6
        self.cellsize_ha = self.cellsize_m2 / 1e4
        self.nodata = self.dem_raster.configs.nodata
        self.rows = self.dem_raster.configs.rows
        self.columns = self.dem_raster.configs.columns

    def create_new_raster(self)->Raster:
        return self.wbe.new_raster(self.dem_raster.configs)

#region Watershed

    @property
    def dem_raster(self)->Raster:
        """
        Original DEM Raster
        """
        return self.get_raster(Names.get_standard_file_name("dem_raster"))

    @property
    def landuse_raster(self)->Raster:
        return self.get_raster(Names.get_standard_file_name("landuse_raster"))
    
    @property
    def soil_raster(self)->Raster:
        return self.get_raster(Names.get_standard_file_name("soil_raster"))
    
    @property
    def stream_network_user_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("stream_shapefile"))  
    
    @property
    def boundary_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("boundary_shapefile"))
    
    @property
    def outlet_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("outlet_shapefile"))
  
    @property
    def farm_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("farm_shapefile"))  
    
    @property
    def field_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("field_shapefile")) 

#endregion

#region Lookup

    @property
    def soil_lookup_csv(self):
        return self.find_file(Names.get_standard_file_name("soil_lookup"))       
    
    @property
    def landuse_lookup_csv(self):
        return self.find_file(Names.get_standard_file_name("landuse_lookup"))       

#endregion 

#region Reach BMP

    @property
    def reach_bmp_vectors(self)->list:
        return [self.point_source_vector, self.flow_diversion_vector, self.reservoir_vector, 
                self.catchbasin_vector, self.grass_waterway_vector, self.access_management_vector,
                self.water_use_vector]

    @property
    def point_source_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("point_source_shapefile")) 

    @property
    def flow_diversion_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("flow_diversion_shapefile")) 

    @property
    def reservoir_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("reservoir_shapefile")) 
    
    @property
    def wetland_boundary_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("wetland_boundary_shapefile"))  
    
    @property
    def wetland_outlet_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("wetland_outlet_shapefile"))       
   
    @property
    def reservoir_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("reservoir_shapefile"))      
   
    @property
    def catchbasin_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("manure_catch_basin_shapefile"))
    
    @property
    def grass_waterway_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("grass_waterway_shapefile"))
    
    @property
    def access_management_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("access_management_shapefile"))
    
    @property
    def water_use_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("water_use_shapefile"))   

#endregion

#region Structure BMP

    @property
    def dugout_boundary_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("dugout_boundary_shapefile"))
    
    @property
    def dugout_outlet_vector(self)->Vector:
        return None
    
    @property
    def wascob_boundary_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("wascob_boundary_shapefile"))
    
    @property
    def wascob_outlet_vector(self)->Vector:
        return None
    
    @property
    def raparian_buffer_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("riparian_buffer_shapefile"))    
    
    @property
    def filter_strip_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("filter_strip_shapefile"))    
    
    @property
    def tile_drain_vector(self)->Vector:
        return self.get_vector(Names.get_standard_file_name("tile_drain_shapefile"))    

#endregion

#region areal non-structure bmp

    @property
    def feedlot_boundary_vector(self)->Vector:
        return self.get_vector(Names.feedlotShpName)
    
    @property
    def feedlot_outlet_vector(self)->Vector:
        return self.get_vector(Names.feedlotOutletShpName)
    
    @property
    def manure_storage_boundary_vector(self)->Vector:
        return self.get_vector(Names.manureStorageShpName)
    
    @property
    def offsite_watering_vector(self)->Vector:
        return self.get_vector(Names.offsiteWinteringShpName)

#endregion

    @property
    def bmp_types(self):
        bmps = []

        #reach bmp
        if self.reservoir_vector is not None:
            bmps.append(BMPType.BMP_TYPE_RESERVOIR)
        if self.flow_diversion_vector is not None:
            bmps.append(BMPType.BMP_TYPE_FLOWDIVERSION_STREAM)
        if self.point_source_vector is not None:
            bmps.append(BMPType.BMP_TYPE_POINTSOURCE)
        if self.wetland_boundary_vector is not None:
            bmps.append(BMPType.BMP_TYPE_WETLAND)
        if self.catchbasin_vector is not None:
            bmps.append(BMPType.BMP_TYPE_MANURE_CATCHBASIN)
        if self.grass_waterway_vector is not None:
            bmps.append(BMPType.BMP_TYPE_GRASSWATERWAY)
        if self.access_management_vector is not None:
            bmps.append(BMPType.BMP_TYPE_ACCESSMGT)
        if self.water_use_vector is not None:
            bmps.append(BMPType.BMP_TYPE_WATERUSE)

        #structure bmp
        if self.dugout_boundary_vector is not None:
            bmps.append(BMPType.BMP_TYPE_DUGOUT)
        if self.wascob_boundary_vector is not None:
            bmps.append(BMPType.BMP_TYPE_WASCOB)
        if self.tile_drain_vector is not None:
            bmps.append(BMPType.BMP_TYPE_FILTERSTRIP)
        if self.raparian_buffer_vector is not None:
            bmps.append(BMPType.BMP_TYPE_RIPARIANBUFFER)
        if self.filter_strip_vector is not None:
            bmps.append(BMPType.BMP_TYPE_FILTERSTRIP)

        #non-structure
        if self.manure_storage_boundary_vector is not None:
            bmps.append(BMPType.BMP_TYPE_MANURE_STORAGE)
        if self.feedlot_boundary_vector is not None:
            bmps.append(BMPType.BMP_TYPE_MANURE_FEEDLOT)

        return bmps  

    def __validate(self):
        """
        Validate necessary files are present.
        Validate all grid inputs so they have same dimension. DEM is used as the standard. 
        Only soil and landuse are used here assuming all other inputs are gnereated from shapefile and would correct dimension as the dem.      
        """

        logger.info("Validating inputs ...")

        #check mandatary files
        if self.dem_raster is None:
            raise ValueError("Can't find DEM")
        
        if self.stream_network_user_vector is None:
            raise ValueError("Can't find user stream network shapefile.")
        
        if self.soil_raster is None:
            raise ValueError("Can't find soil raster")
        
        if self.landuse_raster is None:
            raise ValueError("Can't find landuse raster")       
        
        if self.farm_vector is None:
            raise ValueError("Can't find farm shapefile")       
        
        if self.field_vector is None:
            raise ValueError("Can't find field shapefile")   
        
        if self.soil_lookup_csv is None:
            raise ValueError("Can't find soil lookup")
        
        if self.landuse_lookup_csv is None:
            raise ValueError("Can't find landuse lookup")

        #Check if dem, soil and landuse has the same dimension
        #All input rasters should have the same dimension
        if not RasterExtension.compare_raster_extent(self.dem_raster, self.soil_raster):
            raise ValueError("The soil raster doesn't have same dimension as dem.")

        if not RasterExtension.compare_raster_extent(self.dem_raster, self.landuse_raster):
            raise ValueError("The landuse raster doesn't have same dimension as dem.")
        
        #check manure feedlot, catch basin and storage
        self.__check_manure_feedlot_catchbasin_storage()
        self.__check_reservoir()
        
        #check if the soil/lookup raster ids are included in corresponding lookup csv files
        logger.info("Loading lookup tables ...")
        self.lookup_soil =  Lookup(self.soil_lookup_csv, self.soil_raster)
        self.lookup_landuse  = Lookup(self.landuse_lookup_csv, self.landuse_raster)

    def __check_reservoir(self):
        if self.reservoir_vector is None:
            return 

        ReachBMPReservoir.validate(self.reservoir_vector)


    def __check_manure_feedlot_catchbasin_storage(self):
        """
        check manure feedlot catchbasin and storage.
        Move this logic later to dedicated classes
        
        1. feedlot should have a cb column for catch basin id
        2. manure storage should have feedlot column for source feedlot
        3. feedlot outlet id should match the id in feedlot boundary 
        """

        #feedlot boundary and feedlot outlet
        if self.feedlot_boundary_vector is None:
            return 
        
        #get all feedlot ids
        feedlot_ids = VectorExtension.get_unique_ids(self.feedlot_boundary_vector)     
            
        #make sure feedlot has all required columns
        VectorExtension.check_fields_in_vector(self.feedlot_boundary_vector, Names.fields_feedlot)  
        
        #check feedlot outlet id
        feedlot_ids = VectorExtension.get_unique_ids(self.feedlot_boundary_vector)     
        if self.feedlot_outlet_vector is not None:
            feedlot_outlet_ids = VectorExtension.get_unique_ids(self.feedlot_outlet_vector)

            if not np.array_equal(feedlot_ids.sort(), feedlot_outlet_ids.sort()):
                raise ValueError("The ids in feedlot and catch basin doesn't match. Each feedlot should have only one feedlot outlet. Please check. ")


        #check catch basin column in feedlot
        if self.catchbasin_vector is not None:
            catch_basin_ids = VectorExtension.get_unique_ids(self.catchbasin_vector)
            feedlot_catchbasin_ids = VectorExtension.get_unique_field_value(self.feedlot_boundary_vector, Names.field_name_feedlot_catch_basin)

            for catch_basin in feedlot_catchbasin_ids.values():
                if catch_basin not in catch_basin_ids:
                    raise ValueError(f"Couldn't fine catch basin with id = {catch_basin}. Please check feedlot and catch basin shapefile.")
                
        #check feedlot column in manure storage 
        if self.manure_storage_boundary_vector is not None:
            manure_storage_feedlot_ids = VectorExtension.get_unique_field_value(self.manure_storage_boundary_vector, Names.field_name_feedlot)

            for feedlot in manure_storage_feedlot_ids.values():
                if feedlot not in feedlot_ids:
                    raise ValueError(f"Couldn't fine feedlot with id = {catch_basin}. Please check feedlot and manure storage shapefile.")
            








from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from .bmp_management_base import BMPManagementBaseWithYear
from ...names import Names

class irrigation_management(BMPManagementBaseWithYear):
    """Distribution Table for BMP: Irrigation management (18)"""

    def __init__(self):
        super().__init__()

        self.IrrMon = 1
        self.IrrDay = 1
        self.Days = 2
        self.IrrType = 0
        self.IrrSource = 0
        self.IrrSourceID = 0
        self.IrrRate = 0
        self.IrrMax = 0
        self.IrrEffi = 0
        self.ReturnFlowCo = 0
        self.WstrMax = 0


class irrigation_parameter(BMPTable):
    """Parameter Table for BMP: Irrigation management (18)"""
    __tablename__ = Names.bmp_table_name_irrigation_parameter
    IrrType = Column(Integer,primary_key=True)
    IRRNM = Column(TEXT)
    Year = Column(Integer)
    IrrMon = Column(Integer)
    IrrDay = Column(Integer)
    IrrSource = Column(REAL)
    IrrSourceID = Column(Integer)
    IrrRate = Column(REAL)
    IrrMax = Column(REAL)
    IrrEffi = Column(REAL)
    ReturnFlowCo = Column(REAL)
    WstrMax = Column(REAL)

    def __init__(self):
        self.Year = 1
        self.IrrMon = 5
        self.IrrDay = 10
        self.Days = 1
        self.IrrType = 2
        self.IrrSource = 4
        self.IrrSourceID = 0
        self.IrrRate = 200
        self.IrrMax = 20
        self.IrrEffi = 1
        self.ReturnFlowCo = 0
        self.WstrMax = 0.9



from .bmp import BMP
from whitebox_workflows import Vector, Raster
from .bmp_type import BMPType, ReachBMPColumnNames
from ..database.bmp.reach_bmp import ReachBMPDistribution 
import pandas as pd
from ..database.bmp.bmp_39_offsite_watering import OffsiteWateringParameter

class BMPOffsiteWintering(BMP):
    def __init__(self, bmp_vector:Vector, subbasin_raster:Raster):
        super().__init__(bmp_vector, subbasin_raster)

    @property
    def offsite_wintering_parameters(self):
        return [OffsiteWateringParameter(id,sub) for id,sub in self.subbasins.items()]


from sqlalchemy.orm import DeclarativeBase

class HydroClimateTable(DeclarativeBase):
    def to_dict(self):
        dict = {}
        for col in self.__table__.columns:
            if hasattr(self, col.name):
                dict[col.name] = getattr(self,col.name)

        return dict   


import os
from ..outputs import Outputs
from ..database.hydroclimate.hydroclimate_database import HydroClimateDatabase
from ..names import Names
from .parameter_h5 import ParameterH5
from ..database.bmp.bmp_database import BMPDatabase

class Model:
    """
    Have the model here so all model related structures will be in one place
    """
    def __init__(self, model_folder:str) -> None:
        self.model_folder = model_folder
        self.model_input_folder = os.path.join(self.model_folder, "watershed", "input")
        self.model_output_folder = os.path.join(self.model_folder, "watershed", "output")
        self.model_database_folder = os.path.join(self.model_folder, "database")
        self.__outputs = None
        self.__bmp_database = None
        self.__hydroclimate = None
        self.__parameter_h5 = None

    def create_model_folder(self):
        if not os.path.exists(self.model_input_folder):
            os.makedirs(self.model_input_folder)
        if not os.path.exists(self.model_output_folder):
            os.makedirs(self.model_output_folder)
        if not os.path.exists(self.model_database_folder):
            os.makedirs(self.model_database_folder)

    @property
    def outputs(self)->Outputs:
        if self.__outputs is None:
            self.__outputs = Outputs(self.model_output_folder, self.model_input_folder, self.model_database_folder)
        return self.__outputs
    
    @property
    def bmp_databaes(self)->BMPDatabase:
        if self.__bmp_database is None:
            self.__bmp_database = BMPDatabase(os.path.join(self.model_database_folder,Names.bmpDatabaseName))
        return self.__bmp_database
    
    @property
    def hydroclimate(self)->HydroClimateDatabase:
        if self.__hydroclimate is None:
            self.__hydroclimate = HydroClimateDatabase(os.path.join(self.model_database_folder,Names.hydroclimateDatabasename))
        return self.__hydroclimate
    
    @property
    def parameter_h5(self)->ParameterH5:
        if self.__parameter_h5 is None:
            self.__parameter_h5 = ParameterH5(os.path.join(self.model_output_folder, Names.parameteH5Name))
        return self.__parameter_h5

    def delineate_watershed(self):
        """watershed delineation""" 
        self.outputs.delineate_watershed()

    def generate_parameters(self, reservoir_flow_routing:str, reservoir_flow_data_folder:str):
        """generate parameters"""

        #bmp database
        self.bmp_databaes.create_database_structure()
        self.bmp_databaes.create_spatial_relationship_tables(self.outputs)
        self.bmp_databaes.create_bmp_tables(self.outputs,reservoir_flow_routing, reservoir_flow_data_folder)


    def update_crop_rotation(self,crop_inventory_folder:str, first_year:int, last_year:int):
        """update crop rotation"""
        self.bmp_databaes.update_crop_rotation_AAFC_crop_inventory(crop_inventory_folder, first_year, last_year, self.outputs)

    def run(self, scenario_id, scenario_name):
        pass




from sqlalchemy import Column, Integer, Float, String
from .bmp_table import BMPTable
from ...names import Names

class FieldInfo(BMPTable):
    __tablename__ = Names.bmp_table_name_field_info
    ID = Column(Integer, primary_key=True)
    Area_Ha = Column(Float)

    def __init__(self, id, area_ha):
        self.ID = id
        self.Area_Ha = area_ha


from .bmp_areal import ArealBMP
from ..vector_extension import VectorExtension
from ..database.bmp.bmp_27_manure_storage import ManureStorageParameter, ManureStorageManagement
from ..names import Names

class ArealBMPManureStorage(ArealBMP):
    """Manure Storage"""
    def __init__(self, bmp_vector, flow_direction_raster, subbasin_raster, reach_raster, dem_raster):
        super().__init__(bmp_vector, flow_direction_raster, subbasin_raster, reach_raster, dem_raster)
        
        self.__manure_storage_parameters = None

    @property
    def manure_storage_parameters(self)->list:
        if self.__manure_storage_parameters is None: 
            self.__manure_storage_parameters = []
            feedlots = VectorExtension.get_unique_field_value(self.bmp_vector, Names.field_name_feedlot)            
            for id, subbasin in self.subbasins:
                contribution_area_ha = self.contribution_area_ha[id]
                drainage_fraction = self.contribution_area_subbasin_ratio[id]
                distance_to_reach = self.reach_distance[id]
                height_to_reach = self.reach_elevation_drop[id]

                self.__manure_storage_parameters.append(
                    ManureStorageParameter(id, 
                                           subbasin, 
                                           0 if id not in feedlots else feedlots[id], #feedlot id
                                           contribution_area_ha, 
                                           drainage_fraction, 
                                           distance_to_reach, 
                                           height_to_reach))

        return self.__manure_storage_parameters


import math
from .database.parameter.parameter_database import ParameterDatabase
from .folder_base import FolderBase 
from .names import Names    

class reach_width_depth_parameter:
    def __init__(self, design_storm, parameter_A, parameter_B):
        self.A = parameter_A
        self.B = parameter_B
        self.design_storm = design_storm

class Parameters(FolderBase):
    """all the parameters from parameter database and bmp database. It exposes necessary """
    GRASS_ID = 38

    def __init__(self, database_folder:str)->None:
        super().__init__(database_folder)

        self.parameter_database = ParameterDatabase(self.get_file_path(Names.parameterDatabaseName))
        self.reach_width_parameters = {2: reach_width_depth_parameter(2, 1, 0.56),
                                       10: reach_width_depth_parameter(10, 1.2, 0.56),
                                       100: reach_width_depth_parameter(100, 1.4, 0.56)}
        
        self.reach_depth_parameters = {2: reach_width_depth_parameter(2, 0.04, 0.45),
                                       10: reach_width_depth_parameter(10, 0.12, 0.52),
                                       100: reach_width_depth_parameter(100, 0.18, 0.55)}

    def get_reach_width_parameter(self, design_storm)->reach_width_depth_parameter:
        if design_storm in self.reach_width_parameters:
            return self.reach_width_parameters[design_storm]
        
        return reach_width_depth_parameter(design_storm, 1, 0.56)
    
    def get_reach_depth_parameter(self, design_storm)->reach_width_depth_parameter:
        if design_storm in self.reach_depth_parameters:
            return self.reach_depth_parameters[design_storm]
        
        return reach_width_depth_parameter(design_storm, 0.04, 0.45)

    def get_parameter_lookup(self,parameter_name,parameter_type):
        return self.parameter_database.get_parameter_lookup(parameter_name,parameter_type)

    def get_potential_runoff_coefficient(self, landuse, soil, slope, reach_fraction):
        impervious_ratio = self.parameter_database.get_impervious_ratio(landuse)
        grass_prc = self.parameter_database.get_potential_runoff_coefficient(self.GRASS_ID, soil)
        base_prc = self.parameter_database.get_potential_runoff_coefficient(landuse, soil)
        base_slope = self.parameter_database.get_slope(landuse, soil)

        return self.calculate_potential_runoff_coefficient(
            impervious_ratio, grass_prc, base_prc, base_slope, slope, reach_fraction)
    
    def get_depression_storage_capacity(self, landuse, soil, slope, reach_fraction):
        impervious_ratio = self.parameter_database.get_impervious_ratio(landuse)
        grass_dsc = self.parameter_database.get_depression_storage_capacity(self.GRASS_ID, soil)
        base_dsc = self.parameter_database.get_depression_storage_capacity(landuse, soil)

        return self.calculate_depression_storage_capacity(
            impervious_ratio, grass_dsc, base_dsc, slope, reach_fraction)

    def get_cn2(self, landuse, soil)->float:
        return self.parameter_database.get_cn2(landuse, soil)

    @staticmethod
    def calculate_potential_runoff_coefficient(impervious_ratio, grass_prc, base_prc, base_slope, slope, reach_fraction):
        slope_factor = (slope + base_slope) / (slope + base_slope) if (slope + base_slope) > 0 else 0
        impervious_part = base_prc + (1 - base_prc) * slope_factor
        prc = impervious_part
        
        if 0 < impervious_ratio < 1:  # urban
            pervious_part = grass_prc + (1 - grass_prc) * slope_factor
            prc = impervious_part * impervious_ratio + pervious_part * (1 - impervious_ratio)
        
        if reach_fraction > 0:  # stream
            prc = reach_fraction + (1 - reach_fraction) * prc
        
        prc = max(prc, 0.0)
        prc = min(prc, 1.0)
        
        return prc

    @staticmethod
    def calculate_depression_storage_capacity(impervious_ratio, grass_dsc, base_dsc, slope, reach_surface_fraction):
        slope_factor = math.exp(-9.5 * slope / 100.0)
        dsc = base_dsc * slope_factor
        
        if 0 < impervious_ratio < 1:  # urban
            dsc = (base_dsc * impervious_ratio + grass_dsc * (1 - impervious_ratio)) * slope_factor

        if reach_surface_fraction > 0:  # stream
            dsc = (1 - reach_surface_fraction) * dsc
            dsc = max(dsc, 0.0)
            dsc = min(dsc, 1.0)

        return dsc



import os
from .interpolation import Interpolation
from whitebox_workflows import Raster, WbEnvironment

class ThiessenPolygon(Interpolation):
    def __init__(self, weight_file: str) -> None:
        super().__init__(weight_file)    

    def write_weight_file(self, mask_raster:Raster, station_coordinates:list):    
        wbe = WbEnvironment()

        rows = mask_raster.configs.rows
        cols = mask_raster.configs.columns
        numShapes = len(station_coordinates)
        
        station_id_raster = wbe.new_raster(mask_raster.configs)
        
        dist = [0.0] * numShapes
        rowCount = 0
        
        for row in range(rows):
            for col in range(cols):
                if mask_raster[row, col] > 0:
                    raster_x = mask_raster.get_x_from_column(col)
                    raster_y = mask_raster.get_y_from_row(row)

                    dist[0] = ((station_coordinates[0][0] - raster_x) ** 2 + 
                                (station_coordinates[0][1] - raster_y) ** 2) ** 0.5
                    minDist = dist[0]
                    minDistId = 0
                    
                    for i in range(1, numShapes):
                        dist[i] = ((station_coordinates[i][0] - raster_x) ** 2 + 
                                    (station_coordinates[i][1] - raster_y) ** 2) ** 0.5
                        if dist[i] < minDist:
                            minDist = dist[i]
                            minDistId = i
                    
                    station_id_raster[row, col] = minDistId
        
        sb = []
        sep = os.linesep
        
        sb.append(str(rowCount))
        sb.append(sep)
        sb.append(str(numShapes))
        sb.append(sep)
        
        finalValues = [0] * numShapes
        
        for row in range(rows):
            for col in range(cols):
                currentValue = int(station_id_raster[row, col])
                finalValues = [0] * numShapes
                finalValues[currentValue] = 1
                finalValue = str(finalValues[0])
                for i in range(1, numShapes):
                    finalValue += "\t" + str(finalValues[i])
                
                sb.append(finalValue)
                sb.append(sep)
        
        with open(self.weight_file, 'w') as out:
            out.write(''.join(sb))




from .bmp_reach import ReachBMP
from ..database.bmp.bmp_28_manure_catch_basin import ManureCatchBasinParameter
from whitebox_workflows import Vector, Raster
import pandas as pd

class ReachBMPManureCatchBasin(ReachBMP):
    def __init__(self, bmp_vector:Vector, subbasin_raster:Raster, reach_parameter_df:pd.DataFrame):
        super().__init__(bmp_vector, subbasin_raster)

        self.reach_parameter_df = reach_parameter_df
        self.__manure_catch_basin_parameters = None
        

    @property
    def manure_catch_basin_parameters(self):
        if self.__manure_catch_basin_parameters is None:
            self.__manure_catch_basin_parameters = []

            reach_id_receive_reach_id_dict = {}
            if self.reach_parameter_df is not None:
                reach_id_receive_reach_id_dict = self.reach_parameter_df.set_index("reach_id")["receive_reach_id"].to_dict()
            for id, subbasin in self.subbasins.items():
                receive_reach_id = -1
                if subbasin in reach_id_receive_reach_id_dict:
                    receive_reach_id = reach_id_receive_reach_id_dict[subbasin]
                self.__manure_catch_basin_parameters.append(ManureCatchBasinParameter(id,subbasin,receive_reach_id))

        return self.__manure_catch_basin_parameters


from whitebox_workflows import Raster
import numpy as np
import math
from .interpolation import Interpolation
from ..raster_extension import RasterExtension

class InverseDistance(Interpolation):
    def __init__(self, weight_file: str) -> None:
        super().__init__(weight_file)
    
    def write_weight_file(self, mask_raster:Raster, station_coordinates:list):
        rows = mask_raster.configs.rows
        cols = mask_raster.configs.columns
        numShapes = len(station_coordinates)

        dist = np.zeros(numShapes)
        aValues = np.zeros(numShapes)
        aValuesInt = np.zeros(numShapes)
        rowCount = RasterExtension.get_number_of_valid_cell(mask_raster)

        sb = []
        sb.append(str(rowCount))
        sb.append("\n")
        sb.append(str(numShapes))
        sb.append("\n")

        aValuesIntSum = 0

        for row in range(rows):
            for col in range(cols):
                if mask_raster[row, col] > 0:
                    tempDenom = 0
                    aValuesIntSum = 0
                    raster_x = mask_raster.get_x_from_column(col)
                    raster_y = mask_raster.get_y_from_row(row)

                    for i in range(numShapes):
                        dist[i] = math.sqrt(math.pow(station_coordinates[i][0] - raster_x, 2) +
                                            math.pow(station_coordinates[i][1] - raster_y, 2))
                        tempDenom += 1 / math.pow(dist[i], 2)

                    for i in range(numShapes):
                        tempNum = 1 / math.pow(dist[i], 2)
                        aValues[i] = tempNum / tempDenom
                        aValuesInt[i] = math.floor(aValues[i] * 10000)

                    while aValuesIntSum != 10000:
                        aValuesIntSum = 0
                        for i in range(numShapes):
                            aValuesIntSum += aValuesInt[i]

                        if aValuesIntSum > 10000:
                            tempNums = [aValues[i] * 10000 - math.floor(aValues[i] * 10000) for i in range(numShapes)]
                            for i in range(numShapes):
                                if tempNums[i] < 0.5:
                                    tempNums[i] = 1

                            for i in range(numShapes):
                                if tempNums[i] != 1:
                                    lowestValue = True
                                    for k in range(numShapes):
                                        if tempNums[i] <= tempNums[k]:
                                            lowestValue = True
                                        else:
                                            lowestValue = False
                                            break

                                    if lowestValue:
                                        aValuesInt[i] = aValuesInt[i] - 1
                                        aValues[i] = aValues[i] - 0.0001
                                        break
                        elif aValuesIntSum < 10000:
                            tempNums = [aValues[i] * 10000 - math.floor(aValues[i] * 10000) for i in range(numShapes)]
                            for i in range(numShapes):
                                if tempNums[i] > 0.5:
                                    tempNums[i] = 0

                            for i in range(numShapes):
                                if tempNums[i] != 1:
                                    highestValue = True
                                    for k in range(numShapes):
                                        if tempNums[i] >= tempNums[k]:
                                            highestValue = True
                                        else:
                                            highestValue = False
                                            break

                                    if highestValue:
                                        aValuesInt[i] = aValuesInt[i] + 1
                                        aValues[i] = aValues[i] + 0.0001
                                        break

                        aValuesIntSum = 0
                        for i in range(numShapes):
                            aValuesIntSum += aValuesInt[i]

                    for i in range(numShapes):
                        aValues[i] = aValuesInt[i] / 10000

                    finalValue = str(aValues[0])
                    for i in range(1, numShapes):
                        finalValue += "\t" + str(aValues[i])

                    sb.append(finalValue)
                    sb.append("\n")

        with open(self.weight_file, 'w') as out:
            out.write(''.join(sb))



from typing import Any
from sqlalchemy import Column, Integer, Float, TEXT, REAL, TEXT
from .bmp_table import BMPTable
from ...delineation.structure_attribute import StructureAttribute
import math
from ...names import Names

class Wetland(BMPTable):
    """Parameter Table for BMP: Isolated wetland (9)"""
    __tablename__ = Names.bmp_table_name_wetland
    ID = Column(TEXT, primary_key=True)
    OperationYear = Column(Integer)
    Subbasin = Column(Integer)
    Type = Column(TEXT)
    ContributionArea_ha = Column(REAL)
    NormalArea_ha = Column(REAL)
    NormalVolume_104M3 = Column(REAL)
    MaxArea_ha = Column(REAL)
    MaxVolume_104M3 = Column(REAL)
    K_mmHr = Column(REAL)
    SedimentConEqui_mgL = Column(REAL)
    D50_um = Column(REAL)
    SettVolN_mYr = Column(REAL)
    SettVolP_mYr = Column(REAL)
    ChlaProCo = Column(REAL)
    WaterCalCo = Column(REAL)
    InflowFrac = Column(REAL)
    InitialVolume_104M3 = Column(REAL)
    InitialSediment_mgL = Column(REAL)
    InitialNO3_mgL = Column(REAL)
    InitialNO2_mgL = Column(REAL)
    InitialNH3_mgL = Column(REAL)
    InitialSolP_mgL = Column(REAL)
    InitialOrgN_mgL = Column(REAL)
    InitialOrgP_mgL = Column(REAL)
    RoutingConstant = Column(REAL)
    ClassType = Column(TEXT)

    #use fixed for now
    #read from parameter database later
    co1 = 2.85
    ex1 = 1.22
    co2 = 7.1
    cons = 9.97
    Ivol = 100

    def __init__(self, attribute:StructureAttribute = None):
        if attribute is not None:
            self.ID = attribute.id
            self.ContributionArea_ha = attribute.contribution_area
            self.Subbasin = attribute.subbasin
            self.MaxArea_ha = attribute.area_ha            

            #use the surface area to calculate the volume
            if attribute.area_ha < 70:
                self.MaxVolume_104M3 = self.co1 * math.pow(self.MaxArea_ha, self.ex1) / 10.0
            else:
                self.MaxVolume_104M3 = (self.co2 * self.MaxArea_ha + self.cons) / 10.0

            #set to the same for now
            self.NormalArea_ha = self.MaxArea_ha
            self.NormalVolume_104M3 = self.MaxVolume_104M3

            #initial volume
            self.InitialVolume_104M3 = self.NormalVolume_104M3 * self.Ivol / 100
        else:
            self.ContributionArea_ha = 0
            self.NormalArea_ha = 0
            self.NormalVolume_104M3 = 0
            self.MaxArea_ha = 0
            self.MaxVolume_104M3 = 0
            self.InitialVolume_104M3 = 0
                
        #all wetland is operational
        self.OperationYear = 0
        self.Type = "Isolated"
        
        #these parameters has default values from parameter databaase, wetland table
        self.K_mmHr = 0.05
        self.SedimentConEqui_mgL = 5
        self.D50_um = 10
        self.SettVolN_mYr = 10
        self.SettVolP_mYr = 10
        self.ChlaProCo = 1
        self.WaterCalCo = 1
        self.InflowFrac = 1        
        self.InitialSediment_mgL = 5
        self.InitialNO3_mgL = 0.5
        self.InitialNO2_mgL = 0.1
        self.InitialNH3_mgL = 0.1
        self.InitialSolP_mgL = 0.05
        self.InitialOrgN_mgL = 0.5
        self.InitialOrgP_mgL = 0.05
        self.RoutingConstant = 10


from .config.model_config import ModelConfig
from .config.scenario_config import ScenarioConfig
from .outputs import Outputs

class imWEBs:
    """
    imWEBs main class
    """

    def __init__(self, model_config_file:str, scenario_config_file:str):
        """
        config_file: the imWEBs model configuration file
        """
        self.model_config = ModelConfig(model_config_file)
        self.scenario_config = ScenarioConfig(scenario_config_file)

    def delineate_watershed(self):
        """
        watershed delineation
        """
        self.model_config.delineate_watershed()

    def generate_parameters(self):
        """
        watershed delineation
        """
        self.model_config.generate_parameters()        

    def update_crop_rotation(self):
        """
        update crop rotation
        """
        self.model_config.update_crop_rotation()

    def generate_scenario(self):
        self.scenario_config.generate_model_structure()

    def generate_all(self):
        self.delineate_watershed()
        self.generate_parameters()
        self.update_crop_rotation()
        self.generate_scenario()

        




class TileDrainParameter:
    """Parameter Table for BMP: till drainge management (19)"""

    def __init__(self):
        self.Scenario = -1
        self.Id = -1
        self.StartYear = 1900
        self.StartMon = 1
        self.StartDay = 1
        self.FieldId = -1
        self.OutletReachId = -1
        self.Type = 0
        self.Elevation = -1
        self.Depth = 914.4
        self.ControlDepth = 500
        self.ControlStartMon = 4
        self.ControlEndMon = 10
        self.Radius = 50
        self.Spacing = 12192
        self.OutletCapacity = 0
        self.LagCoefficient = 0.9
        self.DepthToImperviableLayer = 1500
        self.LateralKScale = 1
        self.SedimentCon = 100
        self.OrgNConc = 10
        self.OrgPConc = 10
        self.PRCTile = 0.75
        self.CNTile = 0.75
        self.GWT0 = 300




from whitebox_workflows import WbEnvironment, Raster, Vector
from io import StringIO
import pandas as pd
import logging
import math
from .vector_extension import VectorExtension
from .names import Names
logger = logging.getLogger(__name__)



class RasterExtension:
    @staticmethod
    def get_max_value(raster:Raster)->float:
        """get max value in a raster"""
        rows = raster.configs.rows
        cols = raster.configs.columns
        noddata = raster.configs.nodata
        max_value = float("-inf")
        for row in range(rows):
            for col in range(cols):
                if raster[row, col] != noddata:
                    if max_value < raster[row, col]:
                        max_value = raster[row, col]
        return max_value         

    @staticmethod
    def raster_to_vector(raster:Raster, vector_type:str = "polygon")->Vector:
        """Convert raster to vector and add an id column as the value"""
        wbe = WbEnvironment()

        #there is no need to consider lines
        vector = None
        if vector_type == "point":
            vector = wbe.raster_to_vector_points(raster)
        elif vector_type == "polygon":
            vector = wbe.raster_to_vector_polygons(raster)
        else:
            vector = wbe.raster_to_vector_lines(raster)

        return VectorExtension.add_id_for_raster_value(vector)

    @staticmethod 
    def get_number_of_valid_cell(raster:Raster):
        """Get number of valid cells"""
        rows = raster.configs.rows
        cols = raster.configs.columns
        rowCount = 0
        for row in range(rows):
            for col in range(cols):
                if raster[row, col] > 0:
                    rowCount += 1 #number of valid cell
        return rowCount    

    @staticmethod
    def flow_dir_to_index_delta(flow_dir):
        if  int(flow_dir) not in RasterExtension.flow_dir_to_index_delta:
            raise ValueError(f"{flow_dir} is not an invalid flow direction value.")
        
        delta = RasterExtension.flow_dir_to_index_delta[int(flow_dir)]
        dx = delta[0]
        dy = delta[1]
        return dx, dy

    @staticmethod
    def reclassify(raster:Raster, lookup_dict, mask_raster:Raster = None)->Raster:
        """
        reclassify raster with dictionary. 

        it could be also done with wbe.reclass but have some issues
        """

        wbe = WbEnvironment()
        mapped_raster = wbe.new_raster(raster.configs)
        no_data = raster.configs.nodata

        dict = lookup_dict
        if type(lookup_dict) is list:
            dict = {}
            for item in lookup_dict:
                dict[item[0]] = item[1]

        for row in range(raster.configs.rows):
            for col in range(raster.configs.columns):
                if mask_raster is None or mask_raster[row, col] > 0:
                    old_id = int(raster[row, col])
                    if old_id != no_data:
                        mapped_raster[row,col] = dict[old_id]

        return mapped_raster

    @staticmethod
    def filter_by_values(raster:Raster, values:list)->Raster:
        """
        remove values that are not included in the given value list
        """
        wbe = WbEnvironment()
        filtered_raster = wbe.new_raster(raster.configs)
        no_data = raster.configs.nodata
        for row in range(raster.configs.rows):
            for col in range(raster.configs.columns):
                if raster[row, col] != no_data and raster[row, col] in values:
                    filtered_raster[row, col] = raster[row, col]

        return filtered_raster

    @staticmethod
    def compare_raster_extent(raster1:Raster, raster2:Raster):
        """Check if the raster have same size as the standard raster """
        standar_raster_config = raster1.configs
        check_raster_config = raster2.configs

        return standar_raster_config.rows == check_raster_config.rows and \
            standar_raster_config.columns == check_raster_config.columns and \
            standar_raster_config.resolution_x == check_raster_config.resolution_x and \
            standar_raster_config.resolution_y == check_raster_config.resolution_y and \
            math.fabs(standar_raster_config.east - check_raster_config.east) <= standar_raster_config.resolution_x and \
            math.fabs(standar_raster_config.west == check_raster_config.west) <= standar_raster_config.resolution_y and \
            standar_raster_config.epsg_code == check_raster_config.epsg_code
    
    @staticmethod
    def check_rasters(rasters:dict)->bool:
        """Compare all the rasters in the dictionary and return trun only when all of them has the same resolution and dimension."""
        if len(rasters) <= 1:
            return True
        
        is_same = True
        standard_raster = None
        for key, value in rasters.items():
            if standard_raster is None:
                standard_raster = value
                continue
            
            if not RasterExtension.compare_raster_extent(standard_raster, value):
                raise ValueError(f"The extend of {value.file_name} doesn't match {standard_raster.file_name}")

        return is_same

    @staticmethod
    def get_category_area_ha_dataframe(raster:Raster, area_col_name:str)->pd.DataFrame:
        """Get area of each raster value in ha"""
        wbe = WbEnvironment()
        class_area = wbe.raster_area(raster)
        df = pd.read_csv(StringIO(class_area[1]),skiprows=1, index_col=0, names=[Names.field_name_id,area_col_name])
        df = df / 10000 #convert to ha
        return df
  
    @staticmethod
    def get_zonal_statistics(input_data_raster:Raster, feature_definition_raster:Raster, stat_type, name:str = None)->pd.DataFrame:
        """do zonal statistics and return requested state value as dictionary"""

        wbe = WbEnvironment()

        #do zonal statistics
        stats = wbe.zonal_statistics(input_data_raster,feature_definition_raster)

        #read the result string to dataframe
        df = pd.read_csv(StringIO(stats[1]), sep ="|",  skiprows=5, index_col=1, names=["first","ID","mean","median","min","max","range","stdev","total","last"])
        
        #need to conver the index to the real id
        #this is not necessary for next release
        #it seems this has been fixed
        #df.index = df.index + 1

        if name is not None:
            df[name] = df[stat_type]
            return df[name].to_frame()
        
        return df[stat_type]    
    
    @staticmethod
    def get_majority_count(input_data_raster:Raster, feature_definition_raster:Raster)->dict:
        """Get the majority count of input data raster in each feature definition raster"""

        rows = input_data_raster.configs.rows
        cols = input_data_raster.configs.columns
        nodata_input = input_data_raster.configs.nodata
        nodata_feature = feature_definition_raster.configs.nodata

        majority_count = {}
        for row in range(rows):
            for col in range(cols):
                if input_data_raster[row, col] != nodata_input and feature_definition_raster[row,col] != nodata_feature:
                    input_id = int(input_data_raster[row, col])
                    feature_id = int(feature_definition_raster[row, col])

                    if feature_id not in majority_count:
                        majority_count[feature_id] = {}
                    if input_id not in majority_count[feature_id]:
                        majority_count[feature_id][input_id] = 0
                    
                    majority_count[feature_id][input_id] = majority_count[feature_id][input_id] + 1
        
        feature_counts = {}
        for feature, counts in majority_count.items():
            feature_counts[feature] = max(counts, key = counts.get)

        return feature_counts
    
    @staticmethod
    def get_overlay_raster(spatial1_ras:Raster, spatial2_ras:Raster):
        spatial1_ras_max = spatial1_ras.configs.maximum   
        if spatial1_ras_max == float('-inf') or spatial1_ras_max == float('inf'):
            spatial1_ras_max = RasterExtension.get_max_value(spatial1_ras)   
        raster1_max = int(math.pow(10, int(math.log10(spatial1_ras_max)) + 2))

        #make a new raster with a unique id combining both raster 1 and raster 2
        return spatial1_ras + spatial2_ras * raster1_max, raster1_max

        


from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from .bmp_management_base import BMPManagementBaseWithYear
from ...names import Names

class FertlizerManagement(BMPManagementBaseWithYear):
    """Distribution Table for BMP: Fertilizer management (15)"""

    def __init__(self,FerMon:int,FerDay:int,FerType:int,FerRate:int):
        super().__init__()
        self.FerMon = FerMon
        self.FerDay = FerDay
        self.FerType = FerType
        self.FerRate = FerRate
        self.FerSurface = 0.2

    @staticmethod
    def default_fertilizer_management():
        """
        get default fertilizer management for crop 1-36
        """
        default_fertilizer_fro_crops = {}

        default_fertilizer_fro_crops[1] =[FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 57)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 20)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 90)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 12)],
        default_fertilizer_fro_crops[2] =[FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 1,FerRate = 19)
                ,FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 75)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 13)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 77)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 99)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 10)],
        default_fertilizer_fro_crops[3] =[FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 22)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 75)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 14)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 64)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 56)],
        default_fertilizer_fro_crops[4] =[FertlizerManagement(FerMon = 6,FerDay = 4,FerType = 1,FerRate = 17)
                ,FertlizerManagement(FerMon = 6,FerDay = 4,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 50)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 10)
                ,FertlizerManagement(FerMon = 7,FerDay = 25,FerType = 1,FerRate = 39)
                ,FertlizerManagement(FerMon = 7,FerDay = 25,FerType = 2,FerRate = 9)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 77)],
        default_fertilizer_fro_crops[5] =[FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 1,FerRate = 24)
                ,FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 95)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 14)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 89)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 91)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 10)],
        default_fertilizer_fro_crops[6] =[FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 1,FerRate = 106)
                ,FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 2,FerRate = 15)],
        default_fertilizer_fro_crops[7] =[FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 1,FerRate = 106)
                ,FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 78)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 15)],
        default_fertilizer_fro_crops[8] =[FertlizerManagement(FerMon = 5,FerDay = 21,FerType = 1,FerRate = 17)
                ,FertlizerManagement(FerMon = 5,FerDay = 21,FerType = 2,FerRate = 13)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 85)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 10)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 1,FerRate = 92)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 2,FerRate = 10)],
        default_fertilizer_fro_crops[9] =[FertlizerManagement(FerMon = 4,FerDay = 18,FerType = 1,FerRate = 91)
                ,FertlizerManagement(FerMon = 4,FerDay = 18,FerType = 2,FerRate = 20)
                ,FertlizerManagement(FerMon = 4,FerDay = 25,FerType = 1,FerRate = 11)
                ,FertlizerManagement(FerMon = 4,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 112)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 90)],
        default_fertilizer_fro_crops[11] =[FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 1,FerRate = 57)
                ,FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 2,FerRate = 20)
                ,FertlizerManagement(FerMon = 9,FerDay = 4,FerType = 1,FerRate = 19)
                ,FertlizerManagement(FerMon = 9,FerDay = 4,FerType = 2,FerRate = 10)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 101)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 7)],
        default_fertilizer_fro_crops[12] =[FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 56)],
        default_fertilizer_fro_crops[13] =[FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 90)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 12)],
        default_fertilizer_fro_crops[14] =[FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 20)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 65)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 64)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 73)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 10)],
        default_fertilizer_fro_crops[15] =[FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 1,FerRate = 40)
                ,FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 55)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 16)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 1,FerRate = 63)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 59)],
        default_fertilizer_fro_crops[16] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 73)],
        default_fertilizer_fro_crops[17] =[FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 12)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 54)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 25)],
        default_fertilizer_fro_crops[18] =[FertlizerManagement(FerMon = 7,FerDay = 25,FerType = 1,FerRate = 76)],
        default_fertilizer_fro_crops[19] =[FertlizerManagement(FerMon = 6,FerDay = 11,FerType = 1,FerRate = 18)
                ,FertlizerManagement(FerMon = 6,FerDay = 11,FerType = 2,FerRate = 24)],
        default_fertilizer_fro_crops[20] =[FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 17)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 57)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 1,FerRate = 71)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 2,FerRate = 19)],
        default_fertilizer_fro_crops[21] =[FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 65)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 18)],
        default_fertilizer_fro_crops[23] =[FertlizerManagement(FerMon = 5,FerDay = 14,FerType = 1,FerRate = 6)
                ,FertlizerManagement(FerMon = 5,FerDay = 14,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 1,FerRate = 56)],
        default_fertilizer_fro_crops[24] =[FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 1,FerRate = 36)
                ,FertlizerManagement(FerMon = 5,FerDay = 25,FerType = 2,FerRate = 20)],
        default_fertilizer_fro_crops[25] =[FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 67)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 15)],
        default_fertilizer_fro_crops[26] =[FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 1,FerRate = 25)
                ,FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 2,FerRate = 13)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 68)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 58)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 80)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 11)],
        default_fertilizer_fro_crops[27] =[FertlizerManagement(FerMon = 5,FerDay = 1,FerType = 1,FerRate = 56)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 57)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 1,FerRate = 74)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 101)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 17)],
        default_fertilizer_fro_crops[28] =[FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 1,FerRate = 19)
                ,FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 1,FerRate = 15)
                ,FertlizerManagement(FerMon = 7,FerDay = 18,FerType = 2,FerRate = 20)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 78)],
        default_fertilizer_fro_crops[29] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 90)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 67)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 10)],
        default_fertilizer_fro_crops[30] =[FertlizerManagement(FerMon = 7,FerDay = 25,FerType = 1,FerRate = 56)
                ,FertlizerManagement(FerMon = 7,FerDay = 25,FerType = 2,FerRate = 20)],
        default_fertilizer_fro_crops[31] =[FertlizerManagement(FerMon = 5,FerDay = 18,FerType = 1,FerRate = 22)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 22)],
        default_fertilizer_fro_crops[32] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 54)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 77)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 7)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 10)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 112)],
        default_fertilizer_fro_crops[33] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 67)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 78)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 15)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 20)],
        default_fertilizer_fro_crops[34] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 21)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 1,FerRate = 79)
                ,FertlizerManagement(FerMon = 6,FerDay = 18,FerType = 2,FerRate = 14)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 1,FerRate = 75)
                ,FertlizerManagement(FerMon = 6,FerDay = 25,FerType = 2,FerRate = 12)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 82)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 14)],
        default_fertilizer_fro_crops[35] =[FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 1,FerRate = 6)
                ,FertlizerManagement(FerMon = 5,FerDay = 11,FerType = 2,FerRate = 17)
                ,FertlizerManagement(FerMon = 7,FerDay = 11,FerType = 1,FerRate = 118)],
        default_fertilizer_fro_crops[36] =[FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 1,FerRate = 97)
                ,FertlizerManagement(FerMon = 5,FerDay = 4,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 9,FerDay = 4,FerType = 1,FerRate = 7)
                ,FertlizerManagement(FerMon = 9,FerDay = 4,FerType = 2,FerRate = 11)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 1,FerRate = 35)
                ,FertlizerManagement(FerMon = 9,FerDay = 11,FerType = 2,FerRate = 6)]

        return default_fertilizer_fro_crops
    
class FertilizerParameter(BMPTable):
    __tablename__ = Names.bmp_table_name_fertilizer_parameter
    IFNUM = Column(Integer, primary_key=True)
    FERTNM = Column(TEXT)
    DESCRIPTION = Column(TEXT)
    FMINN = Column(REAL)
    FMINP = Column(REAL)
    FORGN = Column(REAL)
    FORGP = Column(REAL)
    FNH3N = Column(REAL)
    BACTPDB = Column(REAL)
    BACTLPDB = Column(REAL)
    BACKTKDDB = Column(REAL)
    IsManure = Column(Integer)